from amuse.test.amusetest import TestWithMPI, get_path_to_results
import sys
import os.path
from numpy import pi, arange
import numpy.random

from amuse.community.evtwin.interface import EVtwin, EVtwinInterface

from amuse.support.exceptions import AmuseException
from amuse.units import nbody_system, units
from amuse.units.quantities import new_quantity
from amuse.datamodel import Particle, Particles
from amuse.rfi import channel


class TestEVtwinInterface(TestWithMPI):

    def test1(self):
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.cleanup_code()
        self.assertEqual(0, error)
        instance.stop()

    def test2(self):
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)
        error = instance.cleanup_code()
        self.assertEqual(0, error)
        instance.stop()

    def test3(self):
        print("Testing get/set for metallicity...")
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)
        (metallicity, error) = instance.get_metallicity()
        self.assertEqual(0, error)
        self.assertEqual(0.02, metallicity)

        # Z > 0.04 is not allowed
        error = instance.set_metallicity(0.0401)
        self.assertEqual(0, error)
        error = instance.commit_parameters()
        self.assertEqual(-2, error)

        # 0.0 <= Z <= 0.04 is allowed
        any_allowed_Z = numpy.random.random() * 0.04
        error = instance.set_metallicity(any_allowed_Z)
        self.assertEqual(0, error)
        (metallicity, error) = instance.get_metallicity()
        self.assertEqual(0, error)
        self.assertEqual(any_allowed_Z, metallicity)
        error = instance.commit_parameters()
        self.assertEqual(0, error)
        error = instance.cleanup_code()
        self.assertEqual(0, error)
        instance.stop()

    def test4(self):
        print("Testing get/set for maximum number of stars...")
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)
        (value, error) = instance.get_maximum_number_of_stars()
        self.assertEqual(0, error)
        self.assertEqual(10, value)

        error = instance.set_maximum_number_of_stars(10000)
        self.assertEqual(0, error)

        (value, error) = instance.get_maximum_number_of_stars()
        self.assertEqual(0, error)
        self.assertEqual(10000, value)

        error = instance.commit_parameters()
        self.assertEqual(0, error)
        error = instance.cleanup_code()
        self.assertEqual(0, error)
        instance.stop()

    def test5(self):
        print("Testing basic operations (new_particle_method, evolve_one_step etc.)...")
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)
        error = instance.commit_parameters()
        self.assertEqual(0, error)

        (index_of_the_star, error) = instance.new_particle_method(1.05)
        self.assertEqual(0, error)
        self.assertTrue(index_of_the_star >= 0)
        error = instance.commit_particles()
        self.assertEqual(0, error)
        error = instance.set_wind_multiplier(index_of_the_star, 1.0)
        self.assertEqual(0, error)

        (mass, error) = instance.get_mass(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(1.05, mass)

        for i in range(3):
            error = instance.evolve_one_step(index_of_the_star)
            self.assertEqual(0, error)

        (mass, error) = instance.get_mass(index_of_the_star)
        self.assertEqual(0, error)
        self.assertTrue(mass <= 1.05)

        (age, error) = instance.get_age(index_of_the_star)
        self.assertEqual(0, error)
        self.assertTrue(age > 0)

        (x, error) = instance.get_wind_mass_loss_rate(index_of_the_star)
        self.assertEqual(0, error)
        self.assertTrue(x < 1e-13, "mass loss should be less than 1e-13, it is {0}".format(x))
        self.assertTrue(x > 0, "mass loss rate should be more than 0, it is {0}".format(x))

        error = instance.cleanup_code()
        self.assertEqual(0, error)
        instance.stop()

    def test6(self):
        print("Testing EVtwin stop conditions...")
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)

        (value, error) = instance.get_max_age_stop_condition()
        self.assertEqual(0, error)
        self.assertEqual(2.0e12, value)
        for x in range(10, 14):
            error = instance.set_max_age_stop_condition(10 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_max_age_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10 ** x, value)

        (value, error) = instance.get_min_timestep_stop_condition()
        self.assertEqual(0, error)
        self.assertAlmostEqual(1.0e6, value, 5)
        for x in range(-3, 2):
            error = instance.set_min_timestep_stop_condition(10 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_min_timestep_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10 ** x, value)

        instance.stop()

    def test7(self):
        print("Testing EVtwin parameters...")
        instance = EVtwinInterface()
        error = instance.initialize_code()
        self.assertEqual(0, error)
        error = instance.set_ev_path(instance.get_data_directory())
        self.assertEqual(0, error)
        error = instance.commit_parameters()
        self.assertEqual(0, error)

        (value, error) = instance.get_number_of_ionization_elements()
        self.assertEqual(0, error)
        self.assertEqual(5, value)
        for x in range(1, 10):
            error = instance.set_number_of_ionization_elements(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_number_of_ionization_elements()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_convective_overshoot_parameter()
        self.assertEqual(0, error)
        self.assertEqual(0.12, value)
        for x in [0.0, 0.1, 0.12, 0.15]:
            error = instance.set_convective_overshoot_parameter(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_convective_overshoot_parameter()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_mixing_length_ratio()
        self.assertEqual(0, error)
        self.assertEqual(2.0, value)
        for x in [0.1, 1.0, 3.0]:
            error = instance.set_mixing_length_ratio(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_mixing_length_ratio()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_semi_convection_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.04, value)
        for x in [0.0, 0.1]:
            error = instance.set_semi_convection_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_semi_convection_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_thermohaline_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(1.0, value)
        for x in [0.0, 0.5, 1.5]:
            error = instance.set_thermohaline_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_thermohaline_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        instance.stop()

    def test9(self):
        print("Testing adding and removing particles from stellar evolution code...")
        instance = EVtwinInterface()
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_ev_path(instance.get_data_directory()))
        self.assertEqual(0, instance.commit_parameters())

        (indices, errors) = instance.new_particle_method([1.0, 1.0])
        self.assertEqual(errors, [0, 0])
        self.assertEqual(indices, [1, 2])

        self.assertEqual(0, instance.commit_particles())

        for index in indices:
            self.assertEqual(0, instance.evolve_one_step(index))
            (age_after_evolve, error) = instance.get_age(index)
            self.assertEqual(0, error)
            self.assertAlmostEqual(age_after_evolve, 3000000.0, 5)

        self.assertEqual(0, instance.delete_star(1))
        self.assertEqual(instance.get_number_of_particles()['number_of_particles'], 1)
        (age, error) = instance.get_age(1)
        self.assertEqual(error, -1)

        (indices, errors) = instance.new_particle_method([1.0, 1.0])
        self.assertEqual(errors, [0, 0])
        self.assertEqual(indices, [1, 3])
        self.assertEqual(instance.get_number_of_particles()['number_of_particles'], 3)

        for index in indices:
            (age, error) = instance.get_age(index)
            self.assertEqual(0, error)
            print(age, age_after_evolve)
            self.assertTrue(age < age_after_evolve)
            self.assertEqual(0, instance.evolve_one_step(index))
            (age, error) = instance.get_age(index)
            self.assertEqual(0, error)
            self.assertAlmostEqual(age, age_after_evolve)

        instance.stop()


class TestEVtwin(TestWithMPI):

    def test0(self):
        instance = EVtwin()
        instance.initialize_code()
        instance.cleanup_code()
        instance.stop()

    def test1(self):
        print("Testing assigned default parameter values...")
        instance = EVtwin()
        instance.initialize_code()
        instance.parameters.set_defaults()
        self.assertEqual(10, instance.parameters.maximum_number_of_stars)
        instance.parameters.maximum_number_of_stars = 12
        self.assertEqual(12, instance.parameters.maximum_number_of_stars)
        instance.stop()

    def test2(self):
        print("Testing basic operations")
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()

        stars = Particles(1)
        stars.mass = 10 | units.MSun

        instance.particles.add_particles(stars)
        instance.commit_particles()

        self.assertEqual(instance.particles.mass, 10 | units.MSun)
        self.assertAlmostEqual(instance.particles.luminosity, 5695.19757302 | units.LSun, 6)
        self.assertAlmostEqual(instance.particles.radius, 4.07378590 | units.RSun, 6)
        instance.stop()

    def test3(self):
        print("Testing sun recreation")
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()

        stars = Particles(1)
        stars.mass = 1 | units.MSun

        instance.particles.add_particles(stars)
        instance.commit_particles()

        self.assertEqual(instance.particles.mass, 1 | units.MSun)
        self.assertAlmostEqual(instance.particles.luminosity, 0.7098065 | units.LSun, 6)
        self.assertAlmostEqual(instance.particles.radius, 0.8892833 | units.RSun, 6)

        instance.evolve_model(4.8 | units.Gyr)

        self.assertAlmostEqual(instance.particles.mass, 0.999921335 | units.MSun, 6)
        self.assertAlmostEqual(instance.particles.luminosity, 1.04448714 | units.LSun, 6)
        self.assertAlmostEqual(instance.particles.radius, 1.02061451 | units.RSun, 6)

        instance.stop()

    def xtest3(self):
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        stars = Particles(1)

        star = stars[0]
        star.mass = 5.0 | units.MSun
        star.radius = 0.0 | units.RSun

        instance.particles.add_particles(stars)
        instance.commit_particles()

        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()

        previous_type = star.stellar_type
        results = []
        t0 = 0 | units.Myr
        current_time = t0

        while current_time < (115 | units.Myr):
            instance.evolve_model()
            from_code_to_model.copy()

            current_time = star.age
            print((star.age, star.mass, star.stellar_type))
            if not star.stellar_type == previous_type:
                results.append((star.age, star.mass, star.stellar_type))
                previous_type = star.stellar_type

        print(results)
        self.assertEqual(len(results), 6)

        times = (
            104.0 | units.Myr,
            104.4 | units.Myr,
            104.7 | units.Myr,
            120.1 | units.Myr,
            120.9 | units.Myr,
            121.5 | units.Myr
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(result[0].value_in(units.Myr), expected.value_in(units.Myr), 1)

        masses = (
            5.000 | units.MSun,
            5.000 | units.MSun,
            4.998 | units.MSun,
            4.932 | units.MSun,
            4.895 | units.MSun,
            0.997 | units.MSun
        )
        for result, expected in zip(results, masses):
            self.assertAlmostEqual(result[1].value_in(units.MSun), expected.value_in(units.MSun), 3)

        types = (
            "Hertzsprung Gap",
            "First Giant Branch",
            "Core Helium Burning",
            "First Asymptotic Giant Branch",
            "Second Asymptotic Giant Branch",
            "Carbon/Oxygen White Dwarf",
        )

        for result, expected in zip(results, types):
            self.assertEqual(str(result[2]), expected)

        instance.stop()

    def test4(self):
        print("Testing max age stop condition...")
        # masses = [.5, 1.0, 1.5] | units.MSun # Test with fewer particles for speed-up.
        masses = [.5] | units.MSun
        max_age = 9.0 | units.Myr

        number_of_stars = len(masses)
        stars = Particles(number_of_stars)
        for i, star in enumerate(stars):
            star.mass = masses[i]
            star.radius = 0.0 | units.RSun

#       Initialize stellar evolution code
        instance = EVtwin()
        instance.initialize_code()
        instance.parameters.verbosity = True
        if instance.parameters.maximum_number_of_stars < number_of_stars:
            instance.parameters.maximum_number_of_stars = number_of_stars
        self.assertEqual(instance.parameters.max_age_stop_condition, 2e6 | units.Myr)
        instance.parameters.max_age_stop_condition = max_age
        self.assertEqual(instance.parameters.max_age_stop_condition, max_age)
        instance.commit_parameters()
        instance.particles.add_particles(stars)
#       Let the code perform initialization actions after all particles have been created.
        instance.commit_particles()

        from_code_to_model = instance.particles.new_channel_to(stars)

        instance.evolve_model(end_time=8.0 | units.Myr)
        from_code_to_model.copy()

        for i in range(number_of_stars):
            print(stars[i].age.as_quantity_in(units.Myr))
            self.assertTrue(stars[i].age >= 8.0 | units.Myr)
            self.assertTrue(stars[i].age <= max_age)
            self.assertTrue(stars[i].mass <= masses[i])
            self.assertTrue(stars[i].time_step <= max_age)

        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            end_time=2*max_age,
            expected_message=(
                "Error when calling 'evolve_for' of a "
                "'<class 'amuse.community.evtwin.interface.EVtwin'>', "
                "errorcode is 5, error is 'Age greater than maximum "
                "age limit.'"
            )
        )

        instance.stop()

    def test5(self):
        print("Testing adding and removing particles from stellar evolution code...")

        particles = Particles(3)
        particles.mass = 0.3 | units.MSun

        instance = EVtwin()  # redirection="none")
        instance.initialize_code()
        instance.parameters.verbosity = True
        instance.commit_parameters()
        stars = instance.particles
        self.assertEqual(len(stars), 0)  # before creation
        stars.add_particles(particles[:-1])
        instance.commit_particles()
        instance.evolve_model(1.0 | units.Myr)
        self.assertEqual(len(stars), 2)  # before remove
        self.assertAlmostEqual(stars.age, 1.0 | units.Myr)

        stars.remove_particle(particles[0])
        self.assertEqual(len(stars), 1)
        self.assertEqual(instance.get_number_of_particles(), 1)
        instance.evolve_model(2.0 | units.Myr)
        self.assertAlmostEqual(stars[0].age, 2.0 | units.Myr)

        stars.add_particles(particles[::2])
        self.assertEqual(len(stars), 3)  # it's back...
        self.assertAlmostEqual(stars[0].age, 2.0 | units.Myr)
        self.assertAlmostEqual(stars[1].age, 0.0 | units.Myr)
        self.assertAlmostEqual(stars[2].age, 0.0 | units.Myr)  # ... and rejuvenated.

        instance.evolve_model(3.0 | units.Myr)  # The young stars keep their age offset from the old star
        self.assertAlmostEqual(stars.age, [3.0, 1.0, 1.0] | units.Myr)
        instance.evolve_model(4.0 | units.Myr)
        self.assertAlmostEqual(stars.age, [4.0, 2.0, 2.0] | units.Myr)
        instance.stop()

    def test6(self):
        print("Test for obtaining the stellar structure model")
        stars = Particles(2)
        stars.mass = [1.0, 10.0] | units.MSun
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        self.assertEqual(instance.particles.get_number_of_zones(), [199, 199])
        self.assertEqual(len(instance.particles[0].get_radius_profile()), 199)
        self.assertRaises(AmuseException, instance.particles.get_radius_profile,
            expected_message="Querying radius profiles of more than one particle at a time is not supported.")
        self.assertEqual(len(instance.particles[1].get_density_profile()), 199)
        self.assertIsOfOrder(instance.particles[0].get_radius_profile()[-1],          1.0 | units.RSun)
        self.assertIsOfOrder(instance.particles[0].get_temperature_profile()[0],  1.0e7 | units.K)
        self.assertIsOfOrder(instance.particles[0].get_temperature_profile()[-1],  5.0e3 | units.K)
        radius1 = instance.particles[0].get_radius_profile()
        radius2 = radius1[:-1]
        radius2.prepend(0 | units.m)
        delta_radius_cubed = (radius1**3 - radius2**3)
        total_mass = (4./3. * pi * instance.particles[0].get_density_profile() * delta_radius_cubed).sum()
        self.assertAlmostRelativeEqual(total_mass, stars[0].mass, places=1)
        self.assertAlmostEqual(instance.particles[0].get_mu_profile()[:100], [0.62]*100 | units.amu, places=1)
        instance.stop()

    def test7(self):
        print("Test for obtaining the stellar composition structure")
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        self.assertEqual(number_of_zones,    199)
        self.assertEqual(number_of_species,    9)
        self.assertEqual(len(species_names),  number_of_species)
        self.assertEqual(len(composition),    number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(species_names, ['h1', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24', 'si28', 'fe56'])
        self.assertAlmostEqual(composition[0, -1],        0.7, 3)
        self.assertAlmostEqual(composition[1, -1],        0.3 - instance.parameters.metallicity, 3)
        self.assertAlmostEqual(composition[2:, -1].sum(),  instance.parameters.metallicity, 3)
        self.assertAlmostEqual(composition.sum(axis=0), [1.0]*number_of_zones)
        instance.stop()

    def slowtest8(self):
        print("Test for obtaining the stellar composition structure - evolved star")
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model(11.7 | units.Gyr)
        self.assertTrue(instance.particles[0].age >= 11.7 | units.Gyr)
        print(instance.particles[0].stellar_type)
# ~        self.assertTrue(str(instance.particles[0].stellar_type) == "First Giant Branch")
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        self.assertEqual(number_of_zones,    199)
        self.assertEqual(number_of_species,    9)
        self.assertEqual(len(species_names),  number_of_species)
        self.assertEqual(len(composition),    number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(species_names, ['h1', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24', 'si28', 'fe56'])
        self.assertAlmostRelativeEquals(composition[0, -1],        0.7 | units.none, 1)
        self.assertAlmostRelativeEquals(composition[1, -1],        0.3 - instance.parameters.metallicity, 1)
        self.assertAlmostRelativeEquals(composition[2:, -1].sum(),  instance.parameters.metallicity, 1)
        self.assertAlmostEqual(composition.sum(axis=0), [1.0]*number_of_zones | units.none)
        self.assertAlmostEqual(composition[0, 0],        0.00 | units.none)
        self.assertAlmostEqual(composition[1, 0],        1.00 - instance.parameters.metallicity, 3)
        self.assertAlmostEqual(composition[2:, 0].sum(),  instance.parameters.metallicity, 3)
        instance.stop()

    def test9(self):
        print("Test for saving and loading the stellar structure model")
        filenames = ["test1.dump", "test2.dump"]
        filenames = [os.path.join(get_path_to_results(), name) for name in filenames]
        instance = EVtwin()  # redirection="none")
        instance.parameters.verbosity = True
        instance.particles.add_particles(Particles(mass=[0.5, 0.8] | units.MSun))
        instance.evolve_model()
        instance.particles.write_star_to_file(filenames)
        copies = Particles(2)
        copies.filename = filenames
        instance.particles.add_particles(copies)
        instance.evolve_model()
        print(instance.particles)
        import warnings
        warnings.warn("this test's precision has been temporarily (2017) decreased")
        # the reason for this is that the deviation is compiler dependend (and does only happen
        # on some machine/compiler/OS combinations)
        # it may have something to do with initialization of variables in evtwin
        # ~ self.assertAlmostRelativeEquals(instance.particles.temperature, [3644, 4783, 3644, 4783] | units.K, 3)
        self.assertAlmostRelativeEquals(instance.particles.temperature, [3644, 4783, 3644, 4783] | units.K, 2)
        instance.stop()

    def xtest9(self):
        print("Test for changing the stellar structure model (not yet implemented)")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model()

        density_profile = instance.particles[0].get_density_profile()

        self.assertRaises(AmuseException, instance.particles[0].set_density_profile, density_profile[2:],
            expected_message="The length of the supplied vector (197) does not match the number of "
            "mesh zones of the star (199).")

        mass_factor = 1.1
        instance.particles[0].set_density_profile(mass_factor*density_profile)
        self.assertAlmostRelativeEqual(instance.particles[0].get_density_profile(), density_profile*mass_factor, places=10)
        instance.particles.mass *= mass_factor
        instance.evolve_model()

        outer_radius = instance.particles[0].get_radius_profile()
        inner_radius = outer_radius[:-1]
        inner_radius.prepend(0 | units.m)
        delta_radius_cubed = (outer_radius**3 - inner_radius**3)
        integrated_mass = (4./3.*pi*delta_radius_cubed*instance.particles[0].get_density_profile()).sum()
        self.assertAlmostRelativeEqual(integrated_mass, star.mass*mass_factor, places=3)
        instance.stop()
        del instance

    def xtest10(self):
        print("Test for changing the stellar composition (not yet implemented)")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model()

        composition = instance.particles[0].get_chemical_abundance_profiles()
        h1_profile = composition[0] * 1
        he4_profile = composition[1] * 1
        k_surface = -1  # index to the outer mesh cell (surface)

        self.assertAlmostEqual(composition[0, k_surface],  0.7 | units.none, 4)
        self.assertAlmostEqual(composition[1, k_surface],  (0.3 | units.none) - instance.parameters.metallicity, 4)
        self.assertAlmostEqual(composition[2:, k_surface].sum(),  instance.parameters.metallicity, 4)

        composition[0] = he4_profile
        composition[1] = h1_profile
        instance.particles[0].set_chemical_abundance_profiles(composition)
        instance.evolve_model()

        composition = instance.particles[0].get_chemical_abundance_profiles()
        self.assertAlmostEqual(composition[0, k_surface],  (0.3 | units.none) - instance.parameters.metallicity, 4)
        self.assertAlmostEqual(composition[1, k_surface],  0.7 | units.none, 4)
        self.assertAlmostEqual(composition[2:, k_surface].sum(),  instance.parameters.metallicity, 4)
        self.assertAlmostEqual(composition.sum(axis=0), 1.0 | units.none)

        self.assertRaises(AmuseException, instance.particles[0].set_chemical_abundance_profiles, composition[:7],
            expected_message="The length of the supplied vector (7) does not match the number of "
            "chemical species of the star (8).")
        instance.stop()
        del instance

    def test11(self):
        print("Test for importing new stellar models")
        instance = EVtwin()
# ~        instance.parameters.import_model_entropy_force = 1.0
# ~        instance.parameters.import_model_entropy_accuracy = 1.0e-1
        instance.parameters.verbosity = True
        instance.particles.add_particles(Particles(mass=[0.8] | units.MSun))
        instance.evolve_model()

        instance.new_particle_from_model(instance.particles[0].get_internal_structure())
        # The above line is equivalent with:
        # copy = Particles(1)
        # copy.internal_structure = instance.particles[0].get_internal_structure()
        # instance.particles.add_particles(copy)

        number_of_zones = instance.particles[0].get_number_of_zones()
        self.assertEqual(len(instance.particles), 2)
        self.assertEqual(instance.particles[1].get_number_of_zones(), number_of_zones)
        self.assertIsOfOrder(instance.particles[1].get_radius_profile()[-1],          1.0 | units.RSun)
        self.assertIsOfOrder(instance.particles[1].get_temperature_profile()[0],  1.0e7 | units.K)
        self.assertIsOfOrder(instance.particles[1].get_pressure_profile()[0],      1.0e17 | units.barye)

        instance.evolve_model(keep_synchronous=False)
        self.assertAlmostRelativeEqual(
            instance.particles[0].temperature, instance.particles[1].temperature, 2)
        self.assertAlmostRelativeEqual(
            instance.particles[0].luminosity, instance.particles[1].luminosity, 2)
        instance.stop()

    def slowtest11(self):
        print("Test for importing new stellar models, also check long-term evolution")
        instance = EVtwin()
        instance.parameters.verbosity = True
        instance.particles.add_particles(Particles(mass=[0.8] | units.MSun))
        instance.evolve_model()

        copy = Particles(1)
        copy.internal_structure = instance.particles[0].get_internal_structure()
        instance.particles.add_particles(copy)

        number_of_zones = instance.particles[0].get_number_of_zones()
        self.assertEqual(len(instance.particles), 2)
        self.assertEqual(instance.particles[1].get_number_of_zones(), number_of_zones)
        self.assertIsOfOrder(instance.particles[1].get_radius_profile()[-1],          1.0 | units.RSun)
        self.assertIsOfOrder(instance.particles[1].get_temperature_profile()[0],  1.0e7 | units.K)
        self.assertIsOfOrder(instance.particles[1].get_pressure_profile()[0],      1.0e17 | units.barye)

        t1, l1 = simulate_evolution_tracks(instance.particles[0], end_time=26.5 | units.Gyr)
        t2, l2 = simulate_evolution_tracks(instance.particles[1], end_time=26.5 | units.Gyr)
        instance.stop()

        i1 = t1.argmax()  # Maximum temperature ~ turnoff main sequence
        i2 = t2.argmax()
        self.assertAlmostRelativeEqual(t1[i1], t2[i2], 2)
        self.assertAlmostRelativeEqual(l1[i1], l2[i2], 2)
        # plot_tracks(t1, l1, t2, l2)

    def xslowtest11(self):
        print("Test 11: Continue the stellar evolution of a 'merger product' - WIP")
        instance = EVtwin()
        instance.initialize_code()
        instance.commit_parameters()

        instance.parameters.min_timestep_stop_condition = 1.0 | units.s

        stars = Particles(3)
        stars.mass = [1.0, 2.0, 1.0] | units.MSun
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model(1.0 | units.Myr)
        stellar_models = instance.native_stars.get_internal_structure()

        self.assertEqual(len(stellar_models), 3)
        self.assertEqual(len(stellar_models[0]), 199)
        self.assertEqual(len(stellar_models[1]), 199)
        self.assertAlmostEqual(stellar_models[0].mass[198], 1.0 | units.MSun, 2)
        self.assertAlmostEqual(stellar_models[1].mass[198], 2.0 | units.MSun, 2)
        self.assertAlmostEqual(stellar_models[0].mass[0], 0.0 | units.MSun, 2)

        instance.new_particle_from_model(stellar_models[0], instance.particles[0].age)
        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(len(instance.imported_stars), 1)
        imported_stellar_model = instance.imported_stars[0].get_internal_structure()
        self.assertEqual(len(imported_stellar_model), 199)
        self.assertAlmostEqual(imported_stellar_model.mass[198], 1.0 | units.MSun, 2)
        self.assertAlmostEqual(imported_stellar_model.mass[0], 0.0 | units.MSun, 2)
        self.assertAlmostRelativeEqual(imported_stellar_model.X_H, stellar_models[0].X_H, 5)
        self.assertAlmostRelativeEqual(imported_stellar_model.X_He, stellar_models[0].X_He, 5)
        self.assertAlmostRelativeEqual(imported_stellar_model.mass, stellar_models[0].mass, 2)
        self.assertAlmostRelativeEqual(imported_stellar_model.radius[1:], stellar_models[0].radius[1:], 2)
#        instance.evolve_model(2.0 | units.Myr)
        print(instance.particles)
        instance.stop()
        del instance

    def xtest12a(self):
        print("Testing basic operations: evolve_one_step and evolve_for")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = EVtwin()
        se_stars = instance.particles.add_particles(stars)
        self.assertAlmostEqual(se_stars.age, [0.0, 0.0] | units.yr)

        for i in range(3):
            se_stars[0].evolve_one_step()
        self.assertAlmostEqual(se_stars.age, [225488.337629, 0.0] | units.yr, 3)
        number_of_steps = 10
        step_size = se_stars[0].age / number_of_steps
        for i in range(1, number_of_steps + 1):
            se_stars[1].evolve_for(step_size)
            self.assertAlmostEqual(se_stars.age, [number_of_steps, i] * step_size)
        print(se_stars)
        self.assertAlmostRelativeEqual(se_stars[0].age,         se_stars[1].age)
        self.assertAlmostRelativeEqual(se_stars[0].luminosity,  se_stars[1].luminosity, 3)
        self.assertAlmostRelativeEqual(se_stars[0].radius,      se_stars[1].radius, 3)
        self.assertAlmostRelativeEqual(se_stars[0].temperature, se_stars[1].temperature, 3)
        instance.stop()

    def xtest12b(self):
        print("Testing basic operations: evolve_one_step and evolve_for (on subset, WIP: Ticket 304)")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = EVtwin()
        se_stars = instance.particles.add_particles(stars)
        self.assertAlmostEqual(se_stars.age, [0.0, 0.0] | units.yr)

        for i in range(3):
            se_stars[0].evolve_one_step()
        self.assertAlmostEqual(se_stars.age, [225488.337629, 0.0] | units.yr, 3)
        number_of_steps = 10
        step_size = se_stars[0].age / number_of_steps
        for i in range(1, number_of_steps + 1):
            se_stars[1:].evolve_for(step_size)
            self.assertAlmostEqual(se_stars.age, [number_of_steps, i] * step_size)
        print(se_stars)
        self.assertAlmostRelativeEqual(se_stars[0].age,         se_stars[1].age)
        self.assertAlmostRelativeEqual(se_stars[0].luminosity,  se_stars[1].luminosity, 3)
        self.assertAlmostRelativeEqual(se_stars[0].radius,      se_stars[1].radius, 3)
        self.assertAlmostRelativeEqual(se_stars[0].temperature, se_stars[1].temperature, 3)
        instance.stop()

    def xtest13(self):
        print("Test evolve_model optional arguments: end_time and keep_synchronous")
        stars = Particles(3)
        stars.mass = [1.0, 2.0, 3.0] | units.MSun
        instance = EVtwin()
        instance.particles.add_particles(stars)

        self.assertAlmostEqual(instance.particles.age, [0.0, 0.0, 0.0] | units.yr)
        self.assertAlmostEqual(instance.particles.time_step, [70465.105509, 6063.68785133, 1876.53255132] | units.yr, 3)

        print("evolve_model without arguments: use shared timestep = 0.99*min(particles.time_step)")
        instance.evolve_model()
        self.assertAlmostEqual(instance.particles.age, 0.99*([1876.53255132, 1876.53255132, 1876.53255132] | units.yr), 3)
        self.assertAlmostEqual(instance.particles.time_step, [70465.105509, 6063.68785133, 1876.53255132] | units.yr, 3)
        self.assertAlmostEqual(instance.model_time, 0.99*1876.53255132 | units.yr, 3)

        print("evolve_model with end_time: take timesteps, until end_time is reached exactly")
        instance.evolve_model(15000 | units.yr)
        self.assertAlmostEqual(instance.particles.age, [15000.0, 15000.0, 15000.0] | units.yr, 3)
        self.assertAlmostEqual(instance.particles.time_step, [84558.1266108, 7276.4254216, 2251.83906159] | units.yr, 3)
        self.assertAlmostEqual(instance.model_time, 15000.0 | units.yr, 3)

        print("evolve_model with keep_synchronous: use non-shared timestep, particle ages will typically diverge")
        instance.evolve_model(keep_synchronous=False)
        self.assertAlmostEqual(instance.particles.age, (15000 | units.yr) + ([84558.1266108, 7276.4254216, 2251.83906159] | units.yr), 3)
        self.assertAlmostRelativeEquals(instance.particles.time_step, [101469.751933, 8731.71050591, 2702.2068739] | units.yr, 1)
        self.assertAlmostEqual(instance.model_time, 15000.0 | units.yr, 3)  # Unchanged!
        instance.stop()

    def test14(self):
        print("Testing EVtwin states")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = EVtwin()

        print("First do everything manually:", end=' ')
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particle(stars[0])
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()
        print("ok")

        print("initialize_code(), commit_parameters(), (re)commit_particles(), "
            "and cleanup_code() should be called automatically:", end=' ')
        instance = EVtwin()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.RGB_wind_setting = -0.5
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particle(stars[0])
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.add_particle(stars[1])
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')
        print("ok")

    def slowtest15(self):
        print("test evolution of 1000 star sampled over flattish IMF")

        number_of_stars = 1000

        from amuse.ic.salpeter import new_salpeter_mass_distribution
        import numpy

        class notsorandom(object):
            def random(self, N):
                return numpy.array(range(N))/(N-1.)

        masses = new_salpeter_mass_distribution(
            number_of_stars,
            mass_min=0.1 | units.MSun,
            mass_max=100.0 | units.MSun,
            alpha=-1.01, random=notsorandom()
        )

        stars = Particles(mass=masses)

        instance = EVtwin()
        instance.parameters.maximum_number_of_stars = number_of_stars
        instance.parameters.min_timestep_stop_condition = .001 | units.s
        instance.particles.add_particles(stars)

        i = 0
        for p in instance.particles:
            print(i, p.mass)
            p.evolve_for(0.1 | units.Myr)
            i += 1

    def xslowtest16(self):
        print("test full evolution of 1000 star sampled over flattish IMF")

        number_of_stars = 1000

        from amuse.ic.salpeter import new_salpeter_mass_distribution
        import numpy

        class notsorandom(object):
            def random(self, N):
                return numpy.array(range(N))/(N-1.)

        masses = new_salpeter_mass_distribution(
            number_of_stars,
            mass_min=0.1 | units.MSun,
            mass_max=100.0 | units.MSun,
            alpha=-1.01, random=notsorandom()
        )

        stars = Particles(mass=masses)

        instance = EVtwin()
        instance.parameters.maximum_number_of_stars = number_of_stars
        instance.particles.add_particles(stars)

        for p in instance.particles:
            p.evolve_for(13.2 | units.Gyr)

    def slowtest17(self):
        """
        We add multiple particles to evtwin and evolve the stars
        individualy. Evtwin crashes on some combinaties of
        star masses and which star is evolved first.
        """
        exceptions = []
        for i, (masses, indices) in enumerate([
                (
                    [1.21372730283, 1.22207032494, 11.21372730283] | units.MSun,
                    (0, 1)
                ),
                (
                    [1.21372730283, 1.22207032494, 1.21372730283] | units.MSun,
                    (0, 1)
                ),
                (
                    [1.21372730283, 1.22207032494, 1.21372730283] | units.MSun,
                    (1, 0)
                ),
                (
                    [1.21372730283, 1.22207032494] | units.MSun,
                    (0, 1)
                ),
                (
                    [1.21372730283, 11.22207032494, 1.21372730283] | units.MSun,
                    (0, 1)
                ),
                (
                    [1.21372730283, 1.21372730283, 1.21372730283] | units.MSun,
                    (0, 1)
                ),
                (
                    [1.21372730283, 1.22207032494, 1.21372730283] | units.MSun,
                    (0, 1)
                ),
                (
                    [0.101, 1.22207032494, 1.21372730283] | units.MSun,
                    (0, 1)
                )
            ]):

            stars = Particles(mass=masses)

            instance = EVtwin()
            instance.particles.add_particles(stars)

            try:
                index_in_indices = 0
                instance.particles[indices[index_in_indices]].evolve_for(0.1 | units.Myr)
                index_in_indices = 1
                instance.particles[indices[index_in_indices]].evolve_for(0.1 | units.Myr)
            except AmuseException as ex:
                exceptions.append([i, masses, indices, index_in_indices])

            instance.stop()

        if len(exceptions) > 0:
            failure_message = ''
            for index, masses, indices, index_in_indices in exceptions:
                failure_message += '[{0}]: error in '.format(index)
                failure_message += 'first' if index_in_indices == 0 else 'second'
                failure_message += ' evolve_for,'
                failure_message += ' index: {0},'.format(indices[index_in_indices])
                failure_message += ' masses (MSun): {0}.\n'.format(masses.value_in(units.MSun))

            self.fail(failure_message)

    def xtest18(self):
        print("Testing EVtwin calculate_core_mass")
        instance = EVtwin()  # redirection="none")
        star = instance.particles.add_particle(Particle(mass=1 | units.MSun))
        instance.evolve_model(0.4 | units.Gyr)  # VERY short, for test speed up
        central_hydrogen_abundance = star.get_chemical_abundance_profiles()[0][0]
        self.assertTrue(central_hydrogen_abundance < 0.68)  # some hydrogen is burned
        self.assertTrue(central_hydrogen_abundance > 0.67)  # ... but not that much yet
        self.assertEqual(star.calculate_core_mass(core_H_abundance_limit=0.67), 0 | units.MSun)
        self.assertAlmostEqual(star.calculate_core_mass(core_H_abundance_limit=0.71), 1 | units.MSun, 1)

        # For test speed up, we use a weird core_H_abundance_limit to define the "hydrogen exhausted core"
        limit = 0.68
        expected_core_mass = 0.0123182798542 | units.MSun
        self.assertAlmostEqual(star.calculate_core_mass(core_H_abundance_limit=limit), expected_core_mass, 3)

        species_names = star.get_names_of_species()
        self.assertEqual(species_names, ['h1', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24', 'si28', 'fe56'])
        h1_core_mass = star.calculate_core_mass(species=["h1"], core_H_abundance_limit=limit)
        he4_core_mass = star.calculate_core_mass(species=["he4"], core_H_abundance_limit=limit)
        c12_core_mass = star.calculate_core_mass(species=["c12"], core_H_abundance_limit=limit)
        n14_core_mass = star.calculate_core_mass(species=["n14"], core_H_abundance_limit=limit)
        o16_core_mass = star.calculate_core_mass(species=["o16"], core_H_abundance_limit=limit)
        ne20_core_mass = star.calculate_core_mass(species=["ne20"], core_H_abundance_limit=limit)
        mg24_core_mass = star.calculate_core_mass(species=["mg24"], core_H_abundance_limit=limit)
        si28_core_mass = star.calculate_core_mass(species=["si28"], core_H_abundance_limit=limit)
        fe56_core_mass = star.calculate_core_mass(species=["fe56"], core_H_abundance_limit=limit)
        metal_core_mass = star.calculate_core_mass(species=["c12", "n14", "o16", "ne20", "mg24", "si28", "fe56"], core_H_abundance_limit=limit)
        instance.stop()
        self.assertAlmostRelativeEqual(h1_core_mass, expected_core_mass*0.68, 2)
        self.assertAlmostRelativeEqual(he4_core_mass, expected_core_mass*0.30, 2)
        self.assertAlmostRelativeEqual(metal_core_mass, expected_core_mass*0.02, 1)
        self.assertAlmostRelativeEqual(expected_core_mass, he4_core_mass + metal_core_mass + h1_core_mass, 7)
        self.assertAlmostRelativeEqual(metal_core_mass, c12_core_mass + n14_core_mass +
            o16_core_mass + ne20_core_mass + mg24_core_mass + si28_core_mass + fe56_core_mass, 7)

    def xtest19(self):
        print("Testing EVtwin central_temperature and central_density")
        instance = EVtwin()
        stars = instance.particles.add_particles(Particles(mass=[0.1, 1, 10] | units.MSun))
        self.assertIsOfOrder(stars.central_temperature, [4e6, 13e6, 31e6] | units.K)
        self.assertIsOfOrder(stars.central_density, [400, 77, 9] | units.g * units.cm**-3)
        instance.stop()

    def test20(self):
        print("Testing EVtwin manual mass transfer rate")
        instance = EVtwin()
        instance.parameters.verbosity = True
        stars = instance.particles.add_particles(Particles(mass=[0.8, 0.8] | units.MSun))
        stars[0].mass_transfer_rate = -1e-8 | units.MSun/units.yr
        instance.evolve_model()
        # NOTE! no mass transfer during the initial step:
        self.assertEqual(stars[0].mass, stars[1].mass)

        age = stars[0].age
        mass = stars[0].mass
        instance.evolve_model()
        self.assertTrue(stars[0].mass < stars[1].mass)
        # NOTE 2! actual mass transfer rate 10x lower:
        self.assertAlmostRelativeEqual((stars[0].mass-mass) / (stars[0].age-age), -1.0e-9 | units.MSun/units.yr, 4)
        instance.stop()


def simulate_evolution_tracks(star, end_time=1 | units.Gyr):
    luminosity_at_time = [] | units.LSun
    temperature_at_time = [] | units.K
    while star.age < end_time:
        luminosity_at_time.append(star.luminosity)
        temperature_at_time.append(star.temperature)
        star.evolve_one_step()
        print(star.age)
    return temperature_at_time, luminosity_at_time


def plot_tracks(temperature1, luminosity1, temperature2, luminosity2):
    from matplotlib import pyplot
    from amuse.plot import loglog, xlabel, ylabel
    pyplot.figure(figsize=(8, 6))
    pyplot.title('Hertzsprung-Russell diagram', fontsize=12)
    loglog(temperature1, luminosity1)
    loglog(temperature2, luminosity2)
    xlabel('Effective Temperature')
    ylabel('Luminosity')
    pyplot.xlim(pyplot.xlim()[::-1])
    pyplot.show()



import os
import sys
import time
from amuse.test.amusetest import TestWithMPI

import numpy
import math
from amuse.units import nbody_system
from amuse.units import units, constants
from amuse import datamodel
from amuse.ic.plummer import new_plummer_model


from amuse.community.kepler_orbiters.interface import Kepler


class TestKeplerOrbiters(TestWithMPI):

    def test0(self):
        instance = Kepler()
        instance.stop()

    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = Kepler(convert_nbody)  # , redirection="none")#, debugger="xterm")
        instance.initialize_code()

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        instance.central_particle.add_particle(stars[0])
        earth = instance.orbiters.add_particle(stars[1])
        instance.commit_particles()

        instance.evolve_model(365 | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)

        position_at_start = earth.position.value_in(units.AU)[0]
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 6)

        instance.evolve_model(365.0 + (365.0 / 2) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        instance.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        instance.cleanup_code()

        instance.stop()



from amuse.test import amusetest
from amuse.units import constants
from amuse.units.nbody_system import *


class TestNbodyUnits(amusetest.TestCase):
    def test1(self):
        convert_nbody = nbody_to_si(1 | units.parsec, 20 | units.MSun)
        y = 1 | mass
        self.assertEqual(str(y), '1 mass')
        y_in_si = convert_nbody.to_si(y)
        y_in_msun = y_in_si.as_quantity_in(units.MSun)
        self.assertEqual(str(y_in_msun), '20.0 MSun')
        y_in_nbody = convert_nbody.to_nbody(y_in_msun)
        self.assertEqual(str(y_in_nbody), '1.0 mass')

    def test2(self):
        convert_nbody = nbody_to_si(1 | units.MSun, 149.5e6 | units.km)
        y = 29800 | units.m / units.s
        y_in_nbody = convert_nbody.to_nbody(y)
        self.assertEqual(str(y_in_nbody.unit), 'length * time**-1')
        self.assertAlmostEqual(y_in_nbody.number, 1.0, 3)

    def test3(self):
        convert_nbody = nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        y = 1 | length * (time**-1)
        y_in_si = convert_nbody.to_si(y)
        # self.assertEqual(str(y_in_nbody.unit), 'length * time**-1')
        self.assertAlmostEqual(y_in_si.number, 29795.4, -1)

    def test4(self):
        convert_nbody = nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        y_in_nbody = convert_nbody.to_nbody(constants.G)
        self.assertEqual(str(y_in_nbody.unit), 'length**3 * mass**-1 * time**-2')
        self.assertAlmostEqual(y_in_nbody.number, 1.0, 9)

    def test5(self):
        convert_nbody = nbody_to_si(149.5e6 | units.km, 1 | units.MSun)
        y = 29800 | units.m / units.s
        y_in_nbody = convert_nbody.to_nbody(y)
        self.assertEqual(str(y_in_nbody.unit), 'length * time**-1')
        self.assertAlmostEqual(y_in_nbody.number, 1.0, 3)

    def test6(self):
        convert_nbody = nbody_to_si(10 | units.kg, 5 | units.m / units. s)
        y = 5 | units.m / units.s
        y_in_nbody = convert_nbody.to_nbody(y)
        self.assertEqual(str(y_in_nbody.unit), 'length * time**-1')
        self.assertAlmostEqual(y_in_nbody.number, 1.0, 3)
        y_in_si = convert_nbody.to_si(y_in_nbody)
        self.assertAlmostEqual(y_in_si.number, 5.0, 3)

    def test7(self):
        convert_nbody = nbody_to_si(1 | units.kg, 1 | units.m / units. s)
        y = 1 | time
        y_in_si = convert_nbody.to_si(y)
        self.assertEqual(str(y_in_si.unit), 's')
        self.assertAlmostEqual(y_in_si.number, 6.6730000000000296e-11, 3)

    def test8(self):
        self.assertTrue(is_nbody_unit(time / length))
        self.assertFalse(is_nbody_unit(units.s / units.m))

    def test9(self):
        convert_nbody = nbody_to_si(1 | units.kg, 1 | units.m / units. s)
        y = 1.0 | units.none
        y_in_nbody = convert_nbody.to_nbody(y)
        y_in_si = convert_nbody.to_si(y)
        self.assertEqual(y_in_nbody, 1.0 | units.none)
        self.assertEqual(y_in_si, 1.0 | units.none)

    def test10(self):
        self.assertEqual((time / length).to_array_of_floats(), [1, 2, -1, 1, 0, 0, 0, 0, 0])
        self.assertEqual((time / (length * mass ** 2)).to_array_of_floats(), [1, 2, -1, 1, -2, 0, 0, 0, 0])



import os.path
import numpy
from amuse.test.amusetest import TestWithMPI

from amuse.community.sphray.interface import SPHRayInterface, SPHRay
from amuse.units import units
from amuse.datamodel import Particles, create_particle_set
from amuse.datamodel import Particle

from amuse.io import read_set_from_file

# default_options = dict(redirection="none")
default_options = {}


class TestSPHRayInterface(TestWithMPI):

    def setUp(self):
        super(TestWithMPI, self).setUp()

    def test1(self):
        print("Test 1: initialization")

        instance = SPHRayInterface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_sphray_data_directory(instance.data_directory))
        self.assertEqual(0, instance.set_sphray_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        print("Test 2: add, commit_particles")
        instance = SPHRayInterface(**default_options)
        self.assertEqual(0, instance.set_sphray_data_directory(instance.data_directory))
        self.assertEqual(0, instance.set_sphray_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_4K")
        mass, hsml, x, y, z, rho, xe, u = self.read_gas_file(input_file)
        number_of_gas_particles = len(x)
        indices, errors = instance.new_gas_particle(mass, hsml, x, y, z, rho, xe, u)
        self.assertEqual(errors, [0]*number_of_gas_particles)
        self.assertEqual(indices, list(range(1, number_of_gas_particles+1)))

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_sources_001.1")
        L, xs, ys, zs, spctype = self.read_src_file(input_file)
        number_of_src_particles = len(xs)
        s_indices, errors = instance.new_src_particle(L, xs, ys, zs, spctype)
        self.assertEqual(errors, [0]*number_of_src_particles)
        self.assertEqual(s_indices, list(range(number_of_gas_particles+1, number_of_src_particles+number_of_gas_particles+1)))

        self.assertEqual(0, instance.commit_particles())
        mass2, hsml2, x2, y2, z2, rho2, xe2, u2, error = instance.get_state_gas(indices)
        self.assertAlmostEqual((mass-mass2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((hsml-hsml2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((x-x2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((y-y2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((z-z2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((rho-rho2), numpy.zeros_like(x), 7)
#        self.assertAlmostEqual((xe-xe2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((u-u2), numpy.zeros_like(x), 7)

        L2, xs2, ys2, zs2, spctype2, error = instance.get_state_src(s_indices)
        self.assertAlmostEqual((L-L2), numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((xs-xs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((ys-ys2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((zs-zs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((spctype-spctype2), numpy.zeros_like(xs), 7)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test3(self):
        print("Test 3: add, commit_particles, setters, remove")
        instance = SPHRayInterface(**default_options)
        self.assertEqual(0, instance.set_sphray_data_directory(instance.data_directory))
        self.assertEqual(0, instance.set_sphray_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_4K")
        mass, hsml, x, y, z, rho, xe, u = self.read_gas_file(input_file)
        number_of_gas_particles = len(x)
        indices, errors = instance.new_gas_particle(mass, hsml, x, y, z, rho, xe, u)
        self.assertEqual(errors, [0]*number_of_gas_particles)
        self.assertEqual(indices, list(range(1, number_of_gas_particles+1)))

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_sources_001.1")
        L, xs, ys, zs, spctype = self.read_src_file(input_file)
        number_of_src_particles = len(xs)
        s_indices, errors = instance.new_src_particle(L, xs, ys, zs, spctype)
        self.assertEqual(errors, [0]*number_of_src_particles)
        self.assertEqual(s_indices, list(range(number_of_gas_particles+1, number_of_src_particles+number_of_gas_particles+1)))

        self.assertEqual(0, instance.commit_particles())
        mass2, hsml2, x2, y2, z2, rho2, xe2, u2, error = instance.get_state_gas(indices)
        self.assertAlmostEqual((mass-mass2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((hsml-hsml2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((x-x2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((y-y2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((z-z2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((rho-rho2), numpy.zeros_like(x), 7)
#        self.assertAlmostEqual((xe-xe2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((u-u2), numpy.zeros_like(x), 7)

        L2, xs2, ys2, zs2, spctype2, error = instance.get_state_src(s_indices)
        self.assertAlmostEqual((L-L2), numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((xs-xs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((ys-ys2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((zs-zs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((spctype-spctype2), numpy.zeros_like(xs), 7)

        error = instance.set_state_gas(100, 1., 2., 3., 4., 5., 6., 7., 8.)
        self.assertEqual(0, error)
        error = instance.set_state_gas(10000, 1., 2., 3., 4., 5., 6., 7., 8.)
        self.assertEqual(-1, error)

        m, h, x, y, z, rho, xe, u, error = instance.get_state_gas(100)
        self.assertEqual(0, error)
        self.assertAlmostEqual(m, 1., 7)
        self.assertAlmostEqual(h, 2., 7)
        self.assertAlmostEqual(x, 3., 7)
        self.assertAlmostEqual(y, 4., 7)
        self.assertAlmostEqual(z, 5., 7)
        self.assertAlmostEqual(rho, 6., 7)
#        self.assertAlmostEqual(xe, 7., 7)
        self.assertAlmostEqual(u, 8., 7)

        error = instance.remove_gas_particle(100)
        self.assertEqual(0, error)
        error = instance.remove_gas_particle(100)
        self.assertEqual(-4, error)

        error = instance.set_state_src(4097, 1., 2., 3., 4., 5.)
        self.assertEqual(0, error)
        error = instance.set_state_src(10000, 1., 2., 3., 4., 5.)
        self.assertEqual(-1, error)

        L, x, y, z, s, error = instance.get_state_src(4097)
        self.assertEqual(0, error)
        self.assertAlmostEqual(L, 1., 7)
        self.assertAlmostEqual(x, 2., 7)
        self.assertAlmostEqual(y, 3., 7)
        self.assertAlmostEqual(z, 4., 7)
        self.assertAlmostEqual(s, 5., 7)

        error = instance.remove_src_particle(4097)
        self.assertEqual(0, error)
        error = instance.remove_src_particle(4097)
        self.assertEqual(-4, error)

        instance.recommit_particles()
        error = instance.remove_gas_particle(100)
        self.assertEqual(-1, error)
        error = instance.remove_gas_particle(4097)
        self.assertEqual(-1, error)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test4(self):
        print("Test 2: set, get time")
        instance = SPHRayInterface(**default_options)
        self.assertEqual(0, instance.set_sphray_data_directory(instance.data_directory))
        self.assertEqual(0, instance.set_sphray_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_4K")
        mass, hsml, x, y, z, rho, xe, u = self.read_gas_file(input_file)
        number_of_gas_particles = len(x)
        indices, errors = instance.new_gas_particle(mass, hsml, x, y, z, rho, xe, u)
        self.assertEqual(errors, [0]*number_of_gas_particles)
        self.assertEqual(indices, list(range(1, number_of_gas_particles+1)))

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_sources_001.1")
        L, xs, ys, zs, spctype = self.read_src_file(input_file)
        number_of_src_particles = len(xs)
        s_indices, errors = instance.new_src_particle(L, xs, ys, zs, spctype)
        self.assertEqual(errors, [0]*number_of_src_particles)
        self.assertEqual(s_indices, list(range(number_of_gas_particles+1, number_of_src_particles+number_of_gas_particles+1)))

        self.assertEqual(0, instance.commit_particles())

        time, err = instance.get_time()
        self.assertEqual(0, err)
        self.assertEqual(0., time)
        err = instance.set_time(123.)
        self.assertEqual(-1, err)   # because we don't know whether it is safe yet
        time, err = instance.get_time()
        instance.stop()
        self.assertEqual(0, err)
        self.assertEqual(123., time)

    def test5(self):
        instance = SPHRayInterface()
        instance.initialize_code()

        for x, l in [('isothermal', 0), ('H_caseA', 1), ('He_caseA', 1)]:
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, l))
            err = getattr(instance, 'set_'+x)(1)
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, 1))
            err = getattr(instance, 'set_'+x)(0)
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, 0))

        for x, l in [('iontempsolver', 2), ('boundary', 0)]:
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, l))
            err = getattr(instance, 'set_'+x)(1)
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, 1))
            err = getattr(instance, 'set_'+x)(0)
            result, err = getattr(instance, 'get_'+x)()
            self.assertEqual((x, result), (x, 0))

        for x, l in [('raynumber', 1000000.), ('boxsize', 13.2), ("defaultspectype", -1.)]:
            result, err = getattr(instance, 'get_'+x)()
            self.assertAlmostEqual(result, l, 6)
            err = getattr(instance, 'set_'+x)(1.)
            result, err = getattr(instance, 'get_'+x)()
            self.assertAlmostEqual(result, 1., 6)
            err = getattr(instance, 'set_'+x)(0)
            result, err = getattr(instance, 'get_'+x)()
            self.assertAlmostEqual(result, 0., 6)

        result, err = instance.get_globalHefraction()
        self.assertEqual(err, 0)
        self.assertEqual(result, 0.)
        err = instance.set_globalHefraction(0.1)
        instance.stop()
        self.assertEqual(err, -2)

    def test6(self):
        print("Test 3: add, commit_particles, setters, remove with velocity")
        instance = SPHRayInterface(**default_options)
        self.assertEqual(0, instance.set_sphray_data_directory(instance.data_directory))
        self.assertEqual(0, instance.set_sphray_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_4K")
        mass, hsml, x, y, z, rho, xe, u, vx, vy, vz = self.read_gas_file_vel(input_file)
        number_of_gas_particles = len(x)
        indices, errors = instance.new_gas_particle(mass, hsml, x, y, z, rho, xe, u, vx, vy, vz)
        self.assertEqual(errors, [0]*number_of_gas_particles)
        self.assertEqual(indices, list(range(1, number_of_gas_particles+1)))

        input_file = os.path.join(os.path.dirname(__file__), "test_sphray_data_sources_001.1")
        L, xs, ys, zs, spctype = self.read_src_file(input_file)
        number_of_src_particles = len(xs)
        s_indices, errors = instance.new_src_particle(L, xs, ys, zs, spctype)
        self.assertEqual(errors, [0]*number_of_src_particles)
        self.assertEqual(s_indices, list(range(number_of_gas_particles+1, number_of_src_particles+number_of_gas_particles+1)))

        self.assertEqual(0, instance.commit_particles())
        mass2, hsml2, x2, y2, z2, rho2, xe2, u2, error = instance.get_state_gas(indices)
        vx2, vy2, vz2, error = instance.get_vel_gas(indices)
        self.assertAlmostEqual((mass-mass2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((hsml-hsml2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((x-x2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((y-y2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((z-z2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((vx-vx2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((vy-vy2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((vz-vz2)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((rho-rho2), numpy.zeros_like(x), 7)
#        self.assertAlmostEqual((xe-xe2), numpy.zeros_like(x), 7)
        self.assertAlmostEqual((u-u2), numpy.zeros_like(x), 7)

        L2, xs2, ys2, zs2, spctype2, error = instance.get_state_src(s_indices)
        self.assertAlmostEqual((L-L2), numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((xs-xs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((ys-ys2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((zs-zs2)/13200., numpy.zeros_like(xs), 7)
        self.assertAlmostEqual((spctype-spctype2), numpy.zeros_like(xs), 7)

        error = instance.set_state_gas(100, 1., 2., 3., 4., 5., 6., 7., 8.)
        self.assertEqual(0, error)
        error = instance.set_vel_gas(100, 9., 10., 11.)
        self.assertEqual(0, error)
        error = instance.set_state_gas(10000, 1., 2., 3., 4., 5., 6., 7., 8.)
        self.assertEqual(-1, error)

        m, h, x, y, z, rho, xe, u, error = instance.get_state_gas(100)
        vx, vy, vz, error = instance.get_vel_gas(100)
        self.assertEqual(0, error)
        self.assertAlmostEqual(m, 1., 7)
        self.assertAlmostEqual(h, 2., 7)
        self.assertAlmostEqual(x, 3., 7)
        self.assertAlmostEqual(y, 4., 7)
        self.assertAlmostEqual(z, 5., 7)
        self.assertAlmostEqual(rho, 6., 7)
#        self.assertAlmostEqual(xe, 7., 7)
        self.assertAlmostEqual(u, 8., 7)
        self.assertAlmostEqual(vx, 9., 7)
        self.assertAlmostEqual(vy, 10., 7)
        self.assertAlmostEqual(vz, 11., 7)

        error = instance.remove_gas_particle(100)
        self.assertEqual(0, error)
        error = instance.remove_gas_particle(100)
        self.assertEqual(-4, error)

        error = instance.set_state_src(4097, 1., 2., 3., 4., 5.)
        self.assertEqual(0, error)
        error = instance.set_state_src(10000, 1., 2., 3., 4., 5.)
        self.assertEqual(-1, error)

        L, x, y, z, s, error = instance.get_state_src(4097)
        self.assertEqual(0, error)
        self.assertAlmostEqual(L, 1., 7)
        self.assertAlmostEqual(x, 2., 7)
        self.assertAlmostEqual(y, 3., 7)
        self.assertAlmostEqual(z, 4., 7)
        self.assertAlmostEqual(s, 5., 7)

        error = instance.remove_src_particle(4097)
        self.assertEqual(0, error)
        error = instance.remove_src_particle(4097)
        self.assertEqual(-4, error)

        instance.recommit_particles()
        error = instance.remove_gas_particle(100)
        self.assertEqual(-1, error)
        error = instance.remove_gas_particle(4097)
        self.assertEqual(-1, error)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def read_gas_file(self, filename):
        p = read_set_from_file(filename, 'amuse')
        mass = p.mass.number
        hsml = p.smoothing_length.number
        x = p.x.number
        y = p.y.number
        z = p.z.number
        rho = p.rho.number
        u = p.internal_energy.number
        xe = numpy.zeros_like(x)
        return mass, hsml, x, y, z, rho, xe, u

    def read_gas_file_vel(self, filename):
        p = read_set_from_file(filename, 'amuse')
        mass = p.mass.number
        hsml = p.smoothing_length.number
        x = p.x.number
        y = p.y.number
        z = p.z.number
        vx = numpy.random.random(len(mass))
        vy = numpy.random.random(len(mass))
        vz = numpy.random.random(len(mass))
        rho = p.rho.number
        u = p.internal_energy.number
        xe = numpy.zeros_like(x)
        return mass, hsml, x, y, z, rho, xe, u, vx, vy, vz

    def read_src_file(self, filename):
        f = open(filename)
        lines = f.readlines()
        f.close()
        L = []
        x = []
        y = []
        z = []
        spctype = []
        for line in lines:
            l = line.split()
            if len(l) == 9:
                L.append(float(l[6]))
                x.append(float(l[0]))
                y.append(float(l[1]))
                z.append(float(l[2]))
                spctype.append(float(l[7]))
        return numpy.array(L), numpy.array(x), numpy.array(y), numpy.array(z), numpy.array(spctype)


class TestSPHRay(TestWithMPI):

    def setUp(self):
        super(TestWithMPI, self).setUp()

    def test0(self):
        print("test1: basic startup and flow")
        instance = SPHRay()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.parameters.box_size = 100 | units.parsec
        self.assertAlmostRelativeEquals(instance.parameters.box_size, 100 | units.parsec, 7)
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')

        self.assertAlmostRelativeEquals(instance.parameters.box_size, 100 | units.parsec, 7)
        instance.cleanup_code()
        instance.stop()

    def test1(self):
        print("test1: adding particles")

        instance = SPHRay()

        gasparts = self.read_gas_file(os.path.join(os.path.dirname(__file__), "test_sphray_data_4K"))
        srcparts = self.read_src_file(os.path.join(os.path.dirname(__file__), "test_sphray_data_sources_001.1"))

        self.assertEqual(len(instance.gas_particles), 0)
        self.assertEqual(len(instance.src_particles), 0)
        instance.gas_particles.add_particles(gasparts)
        instance.src_particles.add_particles(srcparts)
        self.assertEqual(len(instance.gas_particles), len(gasparts))
        self.assertEqual(len(instance.src_particles), len(srcparts))

        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')

        gaspart2 = instance.gas_particles.copy()

        self.assertAlmostRelativeEquals(gasparts.position, gaspart2.position, 6)
        self.assertAlmostRelativeEquals(gasparts.rho, gaspart2.rho, 6)
        self.assertAlmostRelativeEquals(gasparts.u, gaspart2.u, 6)

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("test2: test parameters")
        instance = SPHRay()

        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')

        for par, val in [("isothermal_flag", False),
            ("hydrogen_case_A_flag", True), ("helium_case_A_flag", True)]:
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val, val1)
            setattr(instance.parameters, par, False)
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val1, False)
            setattr(instance.parameters, par, True)
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val1, True)
            setattr(instance.parameters, par, False)
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val1, False)

        for par, val in [
            ("ionization_temperature_solver", 2), ("boundary_condition", 0)]:
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val, val1)
            setattr(instance.parameters, par, 123)
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val1, 123)

        for par, val in [
            ("spectra_file", "./spectra/thermal1e5.cdf")]:
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val, val1)
            setattr(instance.parameters, par, "somefile")
            val1 = getattr(instance.parameters, par)
            self.assertEqual(val1, "somefile")

        for par, val, tval in [("number_of_rays", 1022.69032205 | (units.Myr**-1), 10000 | units.Myr**-1),
                              ("box_size", 13.2 | units.kpc,  10. | units.kpc),
                              ("default_spectral_type", -1., 1.)]:
            val1 = getattr(instance.parameters, par)
            self.assertAlmostRelativeEqual(val, val1, 6)
            setattr(instance.parameters, par, tval)
            val1 = getattr(instance.parameters, par)
            self.assertAlmostRelativeEqual(val1, tval, 6)

        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.stop()

    def test3(self):
        instance = SPHRay()

        instance.src_particles.add_particle(
            Particle(
                luminosity=1 | 1e48 * units.s**-1,
                x=2 | units.m,
                y=3 | units.m,
                z=4 | units.m,
                SpcType=12.3
            )
        )
        self.assertAlmostRelativeEquals(instance.src_particles.luminosity, 1 | 1e48 * units.s**-1, 6)
        self.assertAlmostRelativeEquals(instance.src_particles.x, 2 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.y, 3 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.z, 4 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.SpcType, 12.3, 7)
        instance.stop()

    def test4(self):
        instance = SPHRay()

        instance.gas_particles.add_particle(
            Particle(
                mass=1 | (10**10*units.MSun),
                x=2 | units.kpc,
                y=3 | units.kpc,
                z=4 | units.kpc,
                h_smooth=0.1 | (units.kpc),
                rho=0.5 | ((10**10*units.MSun) / (units.kpc)**3),
                xion=0.01,
                u=0.2 | (10**5 * units.cm/units.s)**2
            )
        )
        instance.src_particles.add_particle(
            Particle(
                luminosity=1 | 1e48 * units.s**-1,
                x=2 | units.m,
                y=3 | units.m,
                z=4 | units.m,
                SpcType=12.3
            )
        )
        instance.commit_particles()
        self.assertAlmostRelativeEquals(instance.src_particles.luminosity, 1 | 1e48 * units.s**-1, 6)
        self.assertAlmostRelativeEquals(instance.src_particles.x, 2 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.y, 3 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.z, 4 | units.m, 7)
        self.assertAlmostRelativeEquals(instance.src_particles.SpcType, 12.3, 7)

        print(instance.src_particles)
        instance.stop()

    def read_gas_file(self, filename):
        p = read_set_from_file(filename, 'amuse')
        mass = p.mass.number
        hsml = p.smoothing_length.number
        x = p.x.number
        y = p.y.number
        z = p.z.number
        rho = p.rho.number
        u = p.internal_energy.number
        xe = numpy.zeros_like(x)
        return create_particle_set(mass=mass | (10**10*units.MSun), h_smooth=hsml | (units.kpc),
            x=x | (units.kpc), y=y | (units.kpc), z=z | (units.kpc), rho=rho | ((10**10*units.MSun) / (units.kpc)**3),
            xion=xe, u=u | (10**5 * units.cm/units.s)**2)

    def read_src_file(self, filename):
        f = open(filename)
        lines = f.readlines()
        f.close()
        L = []
        x = []
        y = []
        z = []
        spctype = []
        for line in lines:
            l = line.split()
            if len(l) == 9:
                L.append(float(l[6]))
                x.append(float(l[0]))
                y.append(float(l[1]))
                z.append(float(l[2]))
                spctype.append(float(l[7]))
        return create_particle_set(luminosity=numpy.array(L) | (10**48 * units.s**-1),
            x=numpy.array(x) | (units.kpc), y=numpy.array(y) | (units.kpc),
                  z=numpy.array(z) | (units.kpc), SpcType=numpy.array(spctype))



import numpy
from amuse.test import amusetest

import os.path

from amuse.community.phigrape.interface import PhiGRAPEInterface, PhiGRAPE
from amuse.community.hermite.interface import Hermite
from amuse.community.bhtree.interface import BHTree
from amuse.community.ph4.interface import ph4
from amuse.community.fi.interface import Fi
from amuse.community.gadget2.interface import Gadget2
from amuse import io
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel


class TestsForTicket208(amusetest.TestCase):

    def _run_addition_removal_test(
            self,
            instance,
            length_unit=nbody_system.length,
            speed_unit=nbody_system.speed,
            mass_unit=nbody_system.mass
        ):

        instance.initialize_code()

        particles = datamodel.Particles(10)
        particles.mass = numpy.arange(1, 11) | mass_unit
        particles.radius = numpy.arange(1, 11) | length_unit
        particles.x = numpy.arange(1, 11) | length_unit
        particles.y = numpy.arange(1, 11) | length_unit
        particles.z = numpy.arange(1, 11) | length_unit
        particles.vx = numpy.arange(1, 11) | speed_unit
        particles.vy = numpy.arange(1, 11) | speed_unit
        particles.vz = numpy.arange(1, 11) | speed_unit

        instance.particles.add_particles(particles)
        instance.commit_particles()

        self.assertEqual(len(instance.particles), 10)
        self.assertAlmostRelativeEquals(instance.particles.mass.as_quantity_in(mass_unit), list(numpy.arange(1, 11)) | mass_unit)

        instance.particles.remove_particle(particles[2])
        instance.particles.remove_particle(particles[5])

        self.assertEqual(len(instance.particles), 8)
        self.assertAlmostRelativeEquals(instance.particles.mass.as_quantity_in(mass_unit), [1, 2, 4, 5, 7, 8, 9, 10] | mass_unit)

        particles_new = datamodel.Particles(1)
        particles_new.mass = 20 | mass_unit
        particles_new.radius = 21 | length_unit
        particles_new.x = 22 | length_unit
        particles_new.y = 23 | length_unit
        particles_new.z = 24 | length_unit
        particles_new.vx = 25 | speed_unit
        particles_new.vy = 26 | speed_unit
        particles_new.vz = 27 | speed_unit

        instance.particles.add_particles(particles_new)
        self.assertEqual(len(instance.particles), 9)

        self.assertAlmostRelativeEquals(instance.particles.mass.as_quantity_in(mass_unit), [1, 2, 4, 5, 7, 8, 9, 10, 20] | mass_unit)
        self.assertAlmostRelativeEquals(instance.particles.x.as_quantity_in(length_unit), [1, 2, 4, 5, 7, 8, 9, 10, 22] | length_unit)

        instance.cleanup_code()
        instance.stop()

    def test1(self):

        instance = Hermite()
        self._run_addition_removal_test(instance)

    def test2(self):

        instance = PhiGRAPE()
        self._run_addition_removal_test(instance)

    def test3(self):

        instance = BHTree()
        self._run_addition_removal_test(instance)

    def test4(self):

        instance = ph4()
        self._run_addition_removal_test(instance)

    def test5(self):
        instance = Fi()
        self._run_addition_removal_test(instance)

    def test6(self):
        length_unit = units.parsec
        speed_unit = units.parsec / units.Myr
        mass_unit = units.MSun
        instance = Gadget2()
        self._run_addition_removal_test(instance, length_unit, speed_unit, mass_unit)



from amuse.test import amusetest

from amuse.support.interface import InCodeComponentImplementation

from amuse.datamodel.indexing import *

from amuse.datamodel.grids import *

import numpy
import inspect
import collections
from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse import datamodel


class TestGrids(amusetest.TestCase):

    def test1(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.mass = 2.0 | units.kg
        self.assertEqual(grid.mass[0][1][2], 2.0 | units.kg)
        self.assertEqual(grid[0][1][2].mass, 2.0 | units.kg)
        self.assertEqual(len(grid.mass), 5)

    def test2(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.mass = units.kg.new_quantity(numpy.arange(5*4*3).reshape(5, 4, 3))
        self.assertEqual(grid.number_of_dimensions(), 3)
        subgrid = grid[1]
        self.assertEqual(subgrid.number_of_dimensions(), 2)
        self.assertEqual(subgrid.mass.number.shape, (4, 3))

    def test3(self):
        grid = datamodel.Grid(5, 4, 3)
        values = numpy.arange(5*4*3).reshape(5, 4, 3)
        grid.mass = units.kg.new_quantity(values)
        self.assertEqual(grid.number_of_dimensions(), 3)

        subgrid = grid[1][2]
        self.assertEqual(subgrid.number_of_dimensions(), 1)
        self.assertEqual(subgrid.mass.number.shape, (3,))
        self.assertTrue(numpy.all(values[1][2] == subgrid.mass.value_in(units.kg)))

    def test4(self):
        grid = datamodel.Grid(5, 4, 3)
        values = numpy.arange(5*4*3).reshape(5, 4, 3)
        grid.mass = units.kg.new_quantity(values)
        self.assertEqual(grid.number_of_dimensions(), 3)

        gridpoint = grid[1][2][1]
        self.assertEqual(gridpoint.mass, 19.0 | units.kg)
        gridpoint = grid[1][2][2]
        self.assertEqual(gridpoint.mass, 20.0 | units.kg)

    def test5(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.add_calculated_attribute("squared_mass", lambda m: m * m, ["mass",])
        values = numpy.arange(5*4*3).reshape(5, 4, 3)
        grid.mass = units.kg.new_quantity(values)
        gridpoint = grid[1][2][1]
        self.assertEqual(gridpoint.mass, 19.0 | units.kg)
        self.assertEqual(gridpoint.squared_mass, (19.0 | units.kg) ** 2)
        subgrid = grid[1][2]
        self.assertTrue(numpy.all(subgrid.squared_mass == ([18.0, 19.0, 20.0] | units.kg) ** 2))

    def test6(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.add_function_attribute("sum_mass", lambda grid, x: grid.mass.sum() + x, lambda grid, gridpoint, x: gridpoint.mass + x)
        values = numpy.arange(5*4*3).reshape(5, 4, 3)
        grid.mass = units.kg.new_quantity(values)
        gridpoint = grid[1][2][1]
        self.assertEqual(gridpoint.mass, 19.0 | units.kg)
        self.assertEqual(gridpoint.sum_mass(2.0 | units.kg), (21.0 | units.kg))
        subgrid = grid[1][2]
        self.assertTrue(numpy.all(subgrid.sum_mass(2 | units.kg) == (18.0 + 19.0 + 20.0 + 2.0 | units.kg)))

    def test7(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.add_vector_attribute("position", ["x", "y", "z"])
        x = numpy.arange(5*4*3).reshape(5, 4, 3)
        y = x + 100.0
        z = x + 200.0
        grid.x = units.m.new_quantity(x)
        grid.y = units.m.new_quantity(y)
        grid.z = units.m.new_quantity(z)
        gridpoint = grid[1][2][1]
        self.assertEqual(gridpoint.position[0], 19 | units.m)
        self.assertEqual(gridpoint.position[1], 119 | units.m)
        self.assertEqual(gridpoint.position[2], 219 | units.m)
        subgrid = grid[1][2]
        self.assertEqual(subgrid.position[1][0], 19 | units.m)
        self.assertEqual(subgrid.position[1][1], 119 | units.m)
        self.assertEqual(subgrid.position[1][2], 219 | units.m)

    def test8(self):
        grid0 = datamodel.Grid(5, 4, 3)
        x = numpy.arange(5*4*3).reshape(5, 4, 3)
        y = x + 100.0
        grid0.x = units.m.new_quantity(x)
        grid0.y = units.m.new_quantity(y)

        grid1 = datamodel.Grid(5, 4, 3)
        x = numpy.arange(5*4*3).reshape(5, 4, 3)
        x = x + 200.0
        y = x + 200.0
        grid1.x = units.m.new_quantity(x)
        grid1.y = units.m.new_quantity(y)

        self.assertTrue(numpy.all(grid0[1][2].x != grid1[1][2].x))

        channel = grid0.new_channel_to(grid1)
        channel.copy_attributes(["x",])

        self.assertTrue(numpy.all(grid0[1][2].x == grid1[1][2].x))
        self.assertTrue(numpy.all(grid0[1][2].y != grid1[1][2].y))

    def test9(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[0][0][0].x, 0.1 | units.m)
        self.assertEqual(grid[0][0][0].y, 0.125 | units.m)
        self.assertEqual(grid[0][0][0].z, 0.25 | units.m)
        self.assertEqual(grid[..., 0, 0].x, [0.1, 0.3, 0.5, 0.7, 0.9] | units.m)
        self.assertEqual(grid[0, 0, ...].z, [0.25, 0.75] | units.m)

        cellsize = grid.cellsize()
        self.assertAlmostRelativeEquals(cellsize[0], 0.2 | units.m)
        self.assertAlmostRelativeEquals(cellsize[1], 0.25 | units.m)
        self.assertAlmostRelativeEquals(cellsize[2], 0.5 | units.m)

    def test11(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        iarray, jarray, karray = grid.indices()
        for i in range(5):
            for j in range(4):
                for k in range(2):
                    self.assertEqual(iarray[i][j][k], i)
                    self.assertEqual(jarray[i][j][k], j)
                    self.assertEqual(karray[i][j][k], k)
        iarray, jarray, karray = grid.indices()
        i = 0
        for j in range(4):
            for k in range(2):
                self.assertEqual(iarray[i][j][k], i)
                self.assertEqual(jarray[i][j][k], j)
                self.assertEqual(karray[i][j][k], k)

        iarray, jarray, karray = grid[..., 0, 0].indices()
        j = 0
        k = 0
        for i in range(5):
            self.assertEqual(iarray[i], i)
            self.assertEqual(jarray[i], j)
            self.assertEqual(karray[i], k)
        iarray, jarray, karray = grid[3, 2, ...].indices()
        i = 3
        j = 2
        for k in range(2):
            self.assertEqual(iarray[k], i)
            self.assertEqual(jarray[k], j)
            self.assertEqual(karray[k], k)
        iarray, jarray, karray = grid[2, ..., 1].indices()
        i = 2
        k = 1
        for j in range(4):
            self.assertEqual(iarray[j], i)
            self.assertEqual(jarray[j], j)
            self.assertEqual(karray[j], k)

    def test12(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[grid.x > 0.4 | units.m].x.shape, (24,))
        self.assertEqual(grid[grid.x > 0.4 | units.m].x, grid.x[grid.x > 0.4 | units.m])
        iarray, jarray, karray = grid.indices()
        self.assertEqual(grid[grid.x > 0.4 | units.m].indices()[0],  iarray[grid.x > 0.4 | units.m])
        self.assertEqual(grid[grid.x > 0.4 | units.m].indices()[1],  jarray[grid.x > 0.4 | units.m])
        self.assertEqual(grid[grid.x > 0.4 | units.m].indices()[2],  karray[grid.x > 0.4 | units.m])

    def test13(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[0].shape, (4, 2))
        self.assertEqual(grid[0][0].shape, (2,))
        self.assertEqual(grid[..., 2, 1].shape, (5,))

    def test14(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[0].x, grid.x[0])
        self.assertEqual(grid[0][1].x, grid.x[0][1])
        self.assertEqual(grid[1][2][1].x, grid.x[1][2][1])
        self.assertEqual(grid[..., 2, 1].x, grid.x[..., 2, 1])
        self.assertEqual(grid[1, ..., 1].x, grid.x[1, ..., 1])
        self.assertEqual(grid[1, 2, ...].x, grid.x[1, 2, ...])
        self.assertEqual(grid[..., 1].x, grid.x[..., 1])
        self.assertEqual(grid[2, ...].x, grid.x[2, ...])
        self.assertEqual(grid[:, 3, :].x, grid.x[:, 3, :])
        self.assertEqual(grid[:, 3, :].y, grid.y[:, 3, :])
        self.assertEqual(grid[:, 3, :].z, grid.z[:, 3, :])

    def test15(self):

        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        nk = nj = ni = 0
        for plane1 in grid:
            nk += 1
            for plane2 in plane1:
                nj += 1
                for plane3 in plane2:
                    ni += 1
        self.assertEqual(nk, 5)
        self.assertEqual(nj, 4 * 5)
        self.assertEqual(ni, 2 * 4 * 5)

    def test16(self):
        grid = datamodel.new_regular_grid((5, 4, 2, 4), [1.0 | units.m, 1.0 | units.m, 1.0 | units.m, 1.0 | units.s], ('x', 'y', 'z', 't'))
        self.assertEqual(grid.shape, (5, 4, 2, 4))
        self.assertEqual(grid.x.shape, (5, 4, 2, 4))
        self.assertAlmostRelativeEquals(grid[1][2][1].x, ([0.3] * 4) | units.m)
        self.assertAlmostRelativeEquals(grid[1][2][1].t, [0.125, 0.375, 0.625, 0.875] | units.s)
        self.assertEqual(grid[0].x, grid.x[0])
        self.assertEqual(grid[0][1].x, grid.x[0][1])
        self.assertEqual(grid[1][2][1].x, grid.x[1][2][1])
        self.assertEqual(grid[1][2][1][2].x, grid.x[1][2][1][2])

    def test17(self):
        grid = datamodel.new_regular_grid((4, 2), [1.0 | units.m, 1.0 | units.m])
        self.assertEqual(grid.shape, (4, 2))
        self.assertEqual(grid.x.shape, (4, 2))
        self.assertAlmostRelativeEquals(grid[1].x, ([0.375] * 2) | units.m)
        self.assertAlmostRelativeEquals(grid[1][1].y, 0.75 | units.m)
        self.assertEqual(grid[0].x, grid.x[0])
        self.assertEqual(grid[0][1].x, grid.x[0][1])

    def test18(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid.shape, (5, 4, 2))
        self.assertEqual(grid[1:2, ...].x.shape, (1, 4, 2))
        self.assertEqual(grid[1:2, ...].shape, (1, 4, 2))
        self.assertEqual(grid[1:2, ...].x, grid.x[1:2, ...])
        self.assertEqual(grid[1:3, ...].x.shape, (2, 4, 2))
        self.assertEqual(grid[1:3, ...].shape, (2, 4, 2))
        self.assertEqual(grid[1:3, ...].x, grid.x[1:3, ...])

    def test19(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[1:3, ...].x, grid.x[1:3, ...])
        self.assertEqual(grid[1:3, 2:3, ...].x, grid.x[1:3, 2:3, ...])
        self.assertEqual(grid[1:3, 2:3, 0:1].x, grid.x[1:3, 2:3, 0:1])
        self.assertEqual(grid[1:3, ..., 0:1].x, grid.x[1:3, ..., 0:1])
        self.assertEqual(grid[..., 0:1].x, grid.x[..., 0:1])
        self.assertEqual(grid[..., 2:3, 0:1].x, grid.x[..., 2:3, 0:1])

    def test20(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[1:3, :, :].x, grid.x[1:3, :, :])
        self.assertEqual(grid[1:3, 2:3, :].x, grid.x[1:3, 2:3, :])
        self.assertEqual(grid[1:3, 2:3, 0:1].x, grid.x[1:3, 2:3, 0:1])
        self.assertEqual(grid[1:3, :, 0:1].x, grid.x[1:3, :, 0:1])
        self.assertEqual(grid[:, :, 0:1].x, grid.x[:, :, 0:1])
        self.assertEqual(grid[:, 2:3, 0:1].x, grid.x[:, 2:3, 0:1])

    def test21(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid[1:3, :, :].copy().x, grid.x[1:3, :, :])
        self.assertEqual(grid[1:3, :, :].copy().shape, (2, 4, 2))
        self.assertEqual(grid[1:3, 2:3, :].copy().x, grid.x[1:3, 2:3, :])
        self.assertEqual(grid[1:3, 2:3, :].copy().shape, (2, 1, 2))
        self.assertEqual(grid[1:3, 2:3, 0:1].copy().x, grid.x[1:3, 2:3, 0:1])
        self.assertEqual(grid[1:3, 2:3, 0:1].copy().shape, (2, 1, 1))
        self.assertEqual(grid[1:3, :, 0:1].copy().x, grid.x[1:3, :, 0:1])
        self.assertEqual(grid[1:3, :, 0:1].copy().shape, (2, 4, 1))
        self.assertEqual(grid[:, :, 0:1].copy().x, grid.x[:, :, 0:1])
        self.assertEqual(grid[:, :, 0:1].copy().shape, (5, 4, 1))
        self.assertEqual(grid[:, 2:3, 0:1].copy().x, grid.x[:, 2:3, 0:1])
        self.assertEqual(grid[:, 2:3, 0:1].copy().shape, (5, 1, 1))

    def test22(self):
        grid1 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        slice1 = grid1[1:3, :, :].copy()
        slice2 = grid2[1:3, :, :]
        slice1.x = -10 | units.m
        channel = slice1.new_channel_to(slice2)
        channel.copy()
        self.assertEqual(grid2.x[1:3], -10 | units.m)
        self.assertEqual(grid2.x[4], grid1.x[4])

    def test23(self):
        grid1 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual(grid1[1:3, ...].shape, (2, 4, 2))
        slice1 = grid1[1:3, ...].copy()
        slice2 = grid2[1:3, ...]
        slice1.x = -10 | units.m
        channel = slice1.new_channel_to(slice2)
        channel.copy()
        self.assertEqual(grid2.x[1:3], -10 | units.m)
        self.assertEqual(grid2.x[4], grid1.x[4])

    def test24(self):
        particle = datamodel.Particle()
        particle.mass = 10 | units.kg

        grid = datamodel.Grid(5, 4, 3)
        grid.mass = 2.0 | units.kg
        grid.nounit = 10
        self.assertEqual(grid.nounit[0][1][2], 10)
        self.assertEqual(grid[0][1][2].nounit, 10)
        self.assertEqual(len(grid.nounit), 5)
        # grid[0][1][0].particle = particle
        # self.assertEquals(grid.mass[0][1][2], 2.0 | units.kg)
        # self.assertEquals(grid[0][1][0].particle, particle)
        # self.assertEquals(grid[0][1][1].particle, None)

    def test25(self):
        grid = datamodel.Grid(5, 4, 3)
        grid.mass = 2.0 | units.kg
        for cell in grid.iter_cells():
            self.assertEqual(cell.mass, 2.0 | units.kg)

    def test26(self):

        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)

        xv, yv, zv = numpy.mgrid[0:5, 0:4, 0:2]
        xv = xv.flatten()
        yv = yv.flatten()
        zv = zv.flatten()
        i = 0
        for cell in grid.iter_cells():
            expected_position = grid[xv[i], yv[i], zv[i]].position
            self.assertEqual(cell.position, expected_position)
            i += 1

    def test27(self):
        grid = datamodel.new_regular_grid((3, 3), [1.0, 1.0] | units.m)
        subgrid1 = grid[0:1, 0:2]
        subgrid2 = grid[0:1, 0:2]
        subgrid3 = grid[0:2, 0:3][0:1, 0:2]
        subgrid4 = grid[0:1, 0:3]

        self.assertTrue(subgrid1 == subgrid2)
        self.assertTrue(subgrid1 == subgrid3)
        self.assertTrue(subgrid2 == subgrid3)
        self.assertFalse(subgrid1 == subgrid4)
        self.assertFalse(subgrid2 == subgrid4)
        self.assertFalse(subgrid3 == subgrid4)

    def test28(self):
        grid = datamodel.Grid(200, 400)
        subgrid1 = grid[1:-1, 1:-1]
        subgrid2 = subgrid1[3:5, 1:399]
        self.assertEqual(subgrid2.shape, (2, 397))

    def test29(self):
        grid = datamodel.Grid(200)
        subgrid1 = grid[1:-1]
        subgrid2 = subgrid1[3:5]
        self.assertEqual(subgrid2.shape, (2,))

    def test30(self):
        grid = datamodel.Grid(200)
        subgrid1 = grid[1:199]
        subgrid2 = subgrid1[3:5]
        self.assertEqual(subgrid2.shape, (2,))

    def test31(self):
        grid = datamodel.Grid(20, 20, 20)
        a = numpy.zeros((20, 20, 20))
        self.assertEqual(a[1].shape, grid[1].shape)
        self.assertEqual(a[1, 2].shape, grid[1, 2].shape)
        self.assertEqual(a[1:5].shape, grid[1:5].shape)
        self.assertEqual(a[2, 1:5].shape, grid[2, 1:5].shape)
        self.assertEqual(a[2, ...].shape, grid[2, ...].shape)
        self.assertEqual(a[..., 3, :].shape, grid[..., 3, :].shape)
        self.assertEqual(a[..., 3:5, :].shape, grid[..., 3:5, :].shape)
        self.assertEqual(a[..., 3:5, :].shape, grid[..., 3:5, :].shape)
        self.assertEqual(a[::, ::2].shape, grid[::, ::2].shape)

    def test32(self):
        grid = datamodel.Grid(200)
        grid.mass = numpy.arange(200)

        self.assertEqual(grid[1].mass, grid.mass[1])
        self.assertEqual(grid[1:-1].mass, grid.mass[1:-1])
        self.assertEqual(grid[-1:1:-1].mass, grid.mass[-1:1:-1])
        self.assertEqual(grid[-1:1:-1][1:-1].mass, grid.mass[-1:1:-1][1:-1])
        self.assertEqual(grid[-1:1:-1][-1:1].mass, grid.mass[-1:1:-1][-1:1])
        self.assertEqual(grid[-1:1:-1][-1:1:-3].mass, grid.mass[-1:1:-1][-1:1:-3])
        self.assertEqual(grid[300:1:-2][-1:5:-3].mass, grid.mass[300:1:-2][-1:5:-3])
        self.assertEqual(grid[100::-2][::3].mass, grid.mass[100::-2][::3])
        self.assertEqual(grid[100::-2][::-3].mass, grid.mass[100::-2][::-3])

    def test32b(self):
        grid = datamodel.Grid(200)
        grid.mass = numpy.arange(200)

        self.assertEqual(grid[::-1].mass, grid.mass[::-1])
        self.assertEqual(grid[10::-1].mass, grid.mass[10::-1])
        self.assertEqual(grid[:100:2].mass, grid.mass[:100:2])
        self.assertEqual(grid[-1::-1].mass, grid.mass[-1::-1])
        self.assertEqual(grid[-1:-300:-1].mass, grid.mass[-1:-300:-1])
        self.assertEqual(grid[300:-300:-1].mass, grid.mass[300:-300:-1])
        self.assertEqual(grid[300:-300:-7].mass, grid.mass[300:-300:-7])

    def test33(self):
        grid = datamodel.Grid(20)
        grid.mass = numpy.zeros((20, 5))
        self.assertEqual(grid[1].mass, numpy.zeros(5))
        grid.mass = numpy.arange(5)
        self.assertEqual(grid[-1].mass, numpy.arange(5))
        subgrid = grid[::2]
        self.assertEqual(subgrid[-1].mass, numpy.arange(5))
        subgrid[1].mass = 5-numpy.arange(5)
        self.assertEqual(subgrid[1].mass, 5-numpy.arange(5))
        self.assertEqual(grid[2].mass, 5-numpy.arange(5))

    def test34(self):
        grid = datamodel.Grid(20)
        grid.mass = numpy.zeros((20, 5)) | units.kg
        self.assertEqual(grid[1].mass, numpy.zeros(5) | units.kg)
        grid.mass = numpy.arange(5) | units.kg
        self.assertEqual(grid[-1].mass, numpy.arange(5) | units.kg)
        subgrid = grid[::2]
        self.assertEqual(subgrid[-1].mass, numpy.arange(5) | units.kg)
        subgrid[1].mass = (5-numpy.arange(5)) | units.kg
        self.assertEqual(subgrid[1].mass, (5-numpy.arange(5)) | units.kg)
        self.assertEqual(grid[2].mass, (5-numpy.arange(5)) | units.kg)

    def test35(self):
        grid = datamodel.Grid(10, 5)
        grid.mass = numpy.zeros((10, 5, 3))
        self.assertEqual(grid[2, 2].mass, numpy.zeros(3))
        grid[::3, ::2].mass = numpy.arange(3)
        self.assertEqual(grid[3, 2].mass, numpy.arange(3))

    def test36(self):
        grid = datamodel.Grid(10)
        grid.mass = numpy.zeros((10, 5, 3))
        self.assertEqual(grid[2].mass, numpy.zeros((5, 3)))
        grid[::3].mass = numpy.ones((5, 3))
        self.assertEqual(grid[3].mass, numpy.ones((5, 3)))

    def test37(self):
        grid = datamodel.Grid(20)
        grid.mass = numpy.zeros((20, 5))
        grid.mass = numpy.arange(5)
        self.assertEqual(grid[-1].mass, numpy.arange(5))
        self.assertEqual(grid.mass.shape, (20, 5))
        subgrid = grid[::2]
        self.assertEqual(subgrid[-1].mass, numpy.arange(5))
        subgrid[1].mass = 5-numpy.arange(5)
        self.assertEqual(subgrid[1].mass, 5-numpy.arange(5))
        self.assertEqual(grid[2].mass, 5-numpy.arange(5))
        cp = subgrid.copy()
        self.assertEqual(cp[1].mass, 5-numpy.arange(5))
        self.assertEqual(cp.mass.shape, (10, 5))
        cp = grid.copy()
        self.assertEqual(cp.mass.shape, (20, 5))
        self.assertEqual(cp[2].mass, 5-numpy.arange(5))
        self.assertEqual(cp[-1].mass, numpy.arange(5))

    def test38(self):
        grid = datamodel.new_cartesian_grid((10,), 1)
        sub = grid[::2]
        self.assertEqual(sub[0].x, 0.5)
        self.assertEqual(sub[(0,)].x, 0.5)

        grid = datamodel.new_cartesian_grid((10, 10), 1)
        sub = grid[::2, ::]
        self.assertEqual(sub[0, 0].x, 0.5)
        self.assertEqual(sub[(0, 1)].y, 1.5)

    def test39(self):
        grid = datamodel.new_cartesian_grid((10, 10), 1)
        sub = grid[3:6, 5:8]
        self.assertEqual(sub[0:-1, 0:-1].x, sub.x[0:-1, 0:-1])
        self.assertEqual(sub[0:-1, -1].x, sub.x[0:-1, -1])
        self.assertEqual(sub[-1, -1].x, sub.x[-1, -1])
        self.assertEqual(sub[-1, -2].x, sub.x[-1, -2])

    def test40(self):
        grid1 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        grid1.m1 = 1
        grid1.m2 = 2
        channel = grid1.new_channel_to(grid2)
        channel.transform(["m3"], lambda x, y: (x,), ["m1", "m2"])
        self.assertEqual(grid2.m3, 1)
        channel.transform(["m3"], lambda x, y: (y,), ["m1", "m2"])
        self.assertEqual(grid2.m3, 2)
        channel.transform(["m3"], lambda x, y: (x+y,), ["m1", "m2"])
        self.assertEqual(grid2.m3, 3)
        channel.transform(["m3", "m4"], lambda x, y: (x+y, 2*x-y), ["m1", "m2"])
        self.assertEqual(grid2.m3, 3)
        self.assertEqual(grid2.m4, 0)

    def test40b(self):
        grid = datamodel.new_regular_grid((50,), [1.0] | units.m)
        for index in [[0], [0, 3, 4], [1, 2, 2], [[2, 3]], [[0, 1], [2, 3]], list(range(50))]:
            i = numpy.array(index)
            self.assertEqual(grid[i].x, grid.x[i])
            self.assertEqual(grid[i][1:].x, grid.x[i][1:])
            self.assertEqual(grid[i][1::2].x, grid.x[i][1::2])

    def test41(self):
        grid = datamodel.new_regular_grid((10, 10), [1.0, 2.0] | units.m)
        for _i, _j in [([0, 1], [2, 3])]:
            i = numpy.array(_i)
            j = numpy.array(_j)
            self.assertEqual(grid[i, j].x, grid.x[i, j])
            self.assertEqual(grid[i, j][1:].x, grid.x[i, j][1:])
            self.assertEqual(grid[i, j][1::2].x, grid.x[i, j][1::2])

    def test42(self):
        grid = datamodel.new_regular_grid((3, 4, 5, 6), [1.0, 2.0, 3.0, 4.0], axes_names="abcd")
        for _i, _j in [([0], [3]), ([0, 1], [2, 3])]:
            i = numpy.array(_i)
            j = numpy.array(_j)
            self.assertEqual(grid[i, j].a, grid.a[i, j])
            self.assertEqual(grid[0, 1, i, j].a, grid.a[0, 1, i, j])
            self.assertEqual(grid[i, j, 0, 0].a, grid.a[i, j, 0, 0])
            self.assertEqual(grid[i, 1, 2, j].a, grid.a[i, 1, 2, j])

    def test43(self):
        grid = datamodel.new_regular_grid((3, 4, 5, 6), [1.0, 2.0, 3.0, 4.0], axes_names="abcd")
        for _i, _j in [([0, 1], [2, 3])]:
            i = numpy.array(_i)
            j = numpy.array(_j)
            subgrid = grid[:, 0, :, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])
            subgrid = grid[:, 0, 2, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])
            subgrid = grid[:, 0, -2, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])

    def test44(self):
        grid = datamodel.new_regular_grid((3, 4, 5, 6), [1.0, 2.0, 3.0, 4.0], axes_names="abcd")
        for _i, _j in [([-2], [-4]), ([0, -1], [-2, 3])]:
            i = numpy.array(_i)
            j = numpy.array(_j)
            subgrid = grid[:, 0, :, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])
            subgrid = grid[:, 0, 2, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])
            subgrid = grid[:, 0, -2, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])

    def test45(self):
        grid = datamodel.new_regular_grid((3, 4, 5, 6), [1.0, 2.0, 3.0, 4.0], axes_names="abcd")
        for _i, _j in [([-1], [-4])]:
            i = numpy.array(_i)
            j = numpy.array(_j)
            subgrid = grid[::2, 0, :, :]
            self.assertEqual(subgrid[i, j].a, subgrid.a[i, j])

    def test46(self):
        grid = datamodel.new_regular_grid((6,), [1.0], axes_names="abcd")
        subgrid = grid[::2]
        self.assertEqual(subgrid[-2].a, subgrid.a[-2])

        grid = datamodel.new_regular_grid((7,), [1.0], axes_names="abcd")
        subgrid = grid[::2]
        self.assertEqual(subgrid[-2].a, subgrid.a[-2])

        grid = datamodel.new_regular_grid((7,), [1.0], axes_names="abcd")
        subgrid = grid[6:0:-2]
        self.assertEqual(subgrid[-2].a, subgrid.a[-2])

    def test47(self):
        grid = datamodel.Grid()
        grid.mass = 12.
        self.assertEqual(grid[...].mass, 12.)
        self.assertEqual(grid.mass, 12.)

    def test48(self):
        p = datamodel.Grid(3)

        p.a1 = 1.
        p.a2 = 1. | units.rad
        p.a3 = 1. | units.deg

        # test all combinations:

        p[0].a1 = 2.
        p[0].a2 = 2.
        p[0].a3 = 2.

        p[1].a1 = 2. | units.rad
        p[1].a2 = 2. | units.rad
        p[1].a3 = 2. | units.rad

        p[2].a1 = 2. | units.deg
        p[2].a2 = 2. | units.deg
        p[2].a3 = 2. | units.deg

        self.assertEqual(p.a1, [2., 2., (2. | units.deg).value_in(units.none)])
        self.assertEqual(p.a2, [2., 2., (2. | units.deg).value_in(units.none)])
        self.assertEqual(p.a3, [(2. | units.rad).in_(units.deg),
                                  (2. | units.rad).in_(units.deg), 2. | units.deg])

    def test50(self):
        grid = datamodel.Grid(3, 2)
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            grid.savepoint((i + 1) * 1.0 | units.s)

        dens = grid.get_timeline_of_attribute("density")
        self.assertEqual(len(dens), 3)
        self.assertEqual(dens[0][1].shape, (3, 2))
        self.assertEqual(dens[1][1].shape, (3, 2))
        self.assertEqual(dens[2][1].shape, (3, 2))
        self.assertEqual(dens[0][0], 1.0 | units.s)
        self.assertEqual(dens[1][0], 2.0 | units.s)
        self.assertEqual(dens[2][0], 3.0 | units.s)

    def test51(self):
        grid = datamodel.Grid(3, 2)
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            grid.savepoint((i + 1) * 1.0 | units.s)

        dens = grid[0:1, :].get_timeline_of_attribute("density")
        self.assertEqual(len(dens), 3)
        self.assertEqual(dens[0][1].shape, (1, 2))
        self.assertEqual(dens[1][1].shape, (1, 2))
        self.assertEqual(dens[2][1].shape, (1, 2))
        self.assertEqual(dens[0][0], 1.0 | units.s)
        self.assertEqual(dens[1][0], 2.0 | units.s)
        self.assertEqual(dens[2][0], 3.0 | units.s)

    def test51(self):
        grid = datamodel.Grid(3, 2)
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            grid.savepoint((i + 1) * 1.0 | units.s)

        time, dens = grid[0:1, :].get_timeline_of_attribute_as_vector("density")
        self.assertEqual(dens.shape, (3, 1, 2))
        self.assertEqual(dens[0].shape, (1, 2))
        self.assertEqual(dens[1].shape, (1, 2))
        self.assertEqual(dens[2].shape, (1, 2))
        self.assertEqual(time[0], 1.0 | units.s)
        self.assertEqual(time[1], 2.0 | units.s)
        self.assertEqual(time[2], 3.0 | units.s)

    def test52(self):
        grid = datamodel.Grid(3, 2)
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            grid.savepoint((i + 1) * 1.0 | units.s)

        time, dens = grid[2, 1].get_timeline_of_attribute_as_vector("density")
        self.assertEqual(dens.shape, (3,))
        self.assertEqual(dens[0], 0 | units.kg/units.m**3)
        self.assertEqual(dens[1], 1 | units.kg/units.m**3)
        self.assertEqual(dens[2], 2 | units.kg/units.m**3)
        self.assertEqual(time[0], 1.0 | units.s)
        self.assertEqual(time[1], 2.0 | units.s)
        self.assertEqual(time[2], 3.0 | units.s)

    def test53(self):
        grid = datamodel.Grid(3, 2)
        subgrid = grid[0:1, :]
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            subgrid.savepoint((i + 1) * 1.0 | units.s)

        time, dens = subgrid.get_timeline_of_attribute_as_vector("density")
        self.assertEqual(dens.shape, (3, 1, 2))
        self.assertEqual(dens[0], 0 | units.kg/units.m**3)
        self.assertEqual(dens[1], 1 | units.kg/units.m**3)
        self.assertEqual(dens[2], 2 | units.kg/units.m**3)
        self.assertEqual(time[0], 1.0 | units.s)
        self.assertEqual(time[1], 2.0 | units.s)
        self.assertEqual(time[2], 3.0 | units.s)

    def test54(self):
        """
        illustrates getting subgrid/gridpoint with history from subgrid with history
        """
        grid = datamodel.Grid(3, 2)
        subgrid = grid[0:1, :]
        for i in range(3):
            grid.density = (i * 1.0) | units.kg/units.m**3
            subgrid.savepoint((i + 1) * 1.0 | units.s)
        # if the gridpoint derives directly from subgrid, its defined on the original
        # grid (which has no history...)
        subsub = subgrid.savepoint((i + 1) * 1.0 | units.s)
        time, dens = subsub[0, 1].get_timeline_of_attribute_as_vector("density")
        self.assertEqual(dens.shape, (3,))
        self.assertEqual(dens[0], 0 | units.kg/units.m**3)
        self.assertEqual(dens[1], 1 | units.kg/units.m**3)
        self.assertEqual(dens[2], 2 | units.kg/units.m**3)
        self.assertEqual(time[0], 1.0 | units.s)
        self.assertEqual(time[1], 2.0 | units.s)
        self.assertEqual(time[2], 3.0 | units.s)


class TestGridFactories(amusetest.TestCase):
    def test1(self):
        grid1 = datamodel.new_cartesian_grid((4, 5), 1.0 | units.m)
        grid2 = datamodel.new_regular_grid((4, 5), [4.0, 5.0] | units.m)
        grid3 = datamodel.new_rectilinear_grid((4, 5), [numpy.arange(5.) | units.m, numpy.arange(6.) | units.m])

        self.assertEqual(grid1.position, grid2.position)
        self.assertEqual(grid2.position, grid3.position)

    def test2(self):
        grid = datamodel.new_rectilinear_grid((10,), (1.*numpy.arange(11),))
        self.assertEqual(grid._axes_cell_boundaries, 1.*numpy.arange(11))
        grid = datamodel.new_regular_grid((10,), [10.])
        self.assertEqual(grid._lengths, [10.])
        grid = datamodel.new_cartesian_grid((10,), 1.)
        self.assertEqual(grid._cellsize, 1.)
        grid = datamodel.new_regular_grid((10, 20,), [10., 15.])
        self.assertEqual(grid._lengths, [10., 15.])

    def test3(self):
        N = 10
        x, y = numpy.indices((N+1, N+1))
        grid = datamodel.new_structured_grid((N, N), [x, y])
        self.assertEqual(grid.shape, (N, N))
        x, y = numpy.indices((N, N))
        x = x+0.5
        y = y+0.5
        self.assertEqual(grid.x, x)
        self.assertEqual(grid.y, y)

    def test4(self):
        N = 2
        x, y, z = numpy.indices((N+1, N+1, 2*N+1))
        grid = datamodel.new_structured_grid((N, N, 2*N), [x, y, z])
        self.assertEqual(grid.shape, (N, N, 2*N))
        x, y, z = numpy.indices((N, N, 2*N))
        x = x+0.5
        y = y+0.5
        z = z+0.5
        self.assertEqual(grid.x, x)
        self.assertEqual(grid.y, y)
        self.assertEqual(grid.z, z)


class TestGridAttributes(amusetest.TestCase):

    def test1(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertAlmostRelativeEquals(grid.get_minimum_position(),  ([0.0, 0.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(grid.get_maximum_position(),  [1.0, 1.0, 1.0] | units.m)
        self.assertAlmostRelativeEquals(grid.get_volume(),  1.0 | units.m ** 3)
        self.assertTrue(grid.contains([0.5, 0.5, 0.5] | units.m))

    def test2(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertTrue(numpy.all(grid.contains([[0.5, 0.5, 0.5], [0.1, 0.1, 0.1]] | units.m)))
        self.assertFalse(numpy.all(grid.contains([[1.1, 0.5, 0.5], [0.1, 1.1, 0.1]] | units.m)))

    def test3(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        points = grid.points()
        self.assertEqual(points.shape, (6, 5, 3, 3))
        self.assertAlmostRelativeEquals(points[0][0][0],  ([0.0, 0.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[1][0][0],  ([0.2, 0.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[1][1][1],  [0.2, 0.25, 0.5] | units.m)
        self.assertAlmostRelativeEquals(points[0][-1][-1],  ([0.0, 1.0, 1.0] | units.m))
        self.assertAlmostRelativeEquals(points[-1][0][-1],  ([1.0, 0.0, 1.0] | units.m))
        self.assertAlmostRelativeEquals(points[-1][-1][0],  ([1.0, 1.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[-1][-1][-1],  ([1.0, 1.0, 1.0] | units.m))

    def test4(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        points = grid.points().reshape([6*5*3, 3])
        connectivity = grid.connectivity()

        self.assertEqual(connectivity.shape, (5, 4, 2, 8))
        first_cell = connectivity[0][0][0]
        self.assertAlmostRelativeEquals(points[first_cell[0]], [0, 0, 0] | units.m)
        self.assertAlmostRelativeEquals(points[first_cell[1]], [0.2, 0, 0] | units.m)
        self.assertAlmostRelativeEquals(points[first_cell[2]],  ([0, 0.25, 0] | units.m))
        self.assertAlmostRelativeEquals(points[first_cell[3]],  ([0.2, 0.25, 0] | units.m))
        self.assertAlmostRelativeEquals(points[first_cell[4]],  ([0.0, 0.0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[first_cell[5]],  ([0.2, 0.0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[first_cell[6]],  ([0.0, 0.25, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[first_cell[7]],  ([0.2, 0.25, 0.5] | units.m))

        self.assertEqual(connectivity[0][0][0], [0, 15,  3, 18,  1, 16, 4, 19])

    def test5(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        points = grid.points().reshape([6*5*3, 3])
        connectivity = grid.connectivity()

        self.assertEqual(connectivity.shape, (5, 4, 2, 8))
        cell = connectivity[0][0][1]
        self.assertAlmostRelativeEquals(points[cell[0]],  ([0.0, 0.0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[1]],  ([0.2, 0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[2]],  ([0, 0.25, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[3]],  ([0.2, 0.25, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[4]],  ([0.0, 0.0, 1.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[5]],  ([0.2, 0.0, 1.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[6]],  ([0.0, 0.25, 1.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[7]],  ([0.2, 0.25, 1.0] | units.m))

        self.assertEqual(connectivity[0][0][0], [0, 15,  3, 18,  1, 16, 4, 19])

    def test6(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        points = grid.points().reshape([6*5*3, 3])
        connectivity = grid.connectivity()

        self.assertEqual(connectivity.shape, (5, 4, 2, 8))
        cell = connectivity[1][1][1]
        self.assertAlmostRelativeEquals(points[cell[0]], ([0.2, 0.25, 0.5] | units.m) + ([0.0, 0.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[1]], ([0.2, 0.25, 0.5] | units.m) + ([0.2, 0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[2]], ([0.2, 0.25, 0.5] | units.m) + ([0, 0.25, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[3]], ([0.2, 0.25, 0.5] | units.m) + ([0.2, 0.25, 0.0] | units.m))
        self.assertAlmostRelativeEquals(points[cell[4]], ([0.2, 0.25, 0.5] | units.m) + ([0.0, 0.0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[5]], ([0.2, 0.25, 0.5] | units.m) + ([0.2, 0.0, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[6]], ([0.2, 0.25, 0.5] | units.m) + ([0.0, 0.25, 0.5] | units.m))
        self.assertAlmostRelativeEquals(points[cell[7]], ([0.2, 0.25, 0.5] | units.m) + ([0.2, 0.25, 0.5] | units.m))

        self.assertEqual(connectivity[0][0][0], [0, 15,  3, 18,  1, 16, 4, 19])

    def test7(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)

        self.assertAlmostRelativeEquals(grid[1][2][3].position, [3, 5, 7] | units.m)

        grid[1][2][3].position = [7, 5, 3] | units.m
        self.assertAlmostRelativeEquals(grid[1][2][3].position, [7, 5, 3] | units.m)

        grid[1][2][3].position += [1, 2, 3] | units.m
        self.assertAlmostRelativeEquals(grid[1][2][3].position, [8, 7, 6] | units.m)

    def test8(self):
        grid = datamodel.new_regular_grid((5, 4), [1.0, 1.0] | units.m)
        self.assertAlmostRelativeEquals(grid.get_minimum_position(),  ([0.0, 0.0] | units.m))
        self.assertAlmostRelativeEquals(grid.get_maximum_position(),  [1.0, 1.0] | units.m)
        self.assertAlmostRelativeEquals(grid.get_volume(),  1.0 | units.m ** 2)
        self.assertTrue(grid.contains([0.5, 0.5] | units.m))

    def test9(self):
        grid = datamodel.new_regular_grid((5, 4, 2), [1.0, 1.0, 1.0] | units.m)
        self.assertEqual((0, 0, 0), grid.get_minimum_index())
        self.assertEqual((4, 3, 1), grid.get_maximum_index())

    def test10(self):
        grid1 = datamodel.new_regular_grid((5, 4), [1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((5, 4), [.1, .1] | units.m)
        grid3 = datamodel.new_regular_grid((5, 4), [.1, .1] | units.m, offset=[0.5, 0.6] | units.m)
        self.assertTrue(grid1.overlaps(grid2))
        self.assertTrue(grid1.overlaps(grid3))
        self.assertFalse(grid2.overlaps(grid3))
        self.assertTrue(grid2.overlaps(grid1))
        self.assertTrue(grid3.overlaps(grid1))
        self.assertFalse(grid3.overlaps(grid2))

    def test11(self):
        grid1 = datamodel.new_regular_grid((4, 4), [1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((4, 4), [1.0, 1.0] | units.m, offset=[-0.5, -0.5] | units.m)
        self.assertTrue(grid1.overlaps(grid2))
        overlap = grid1.get_overlap_with(grid2)
        self.assertEqual(overlap.position, grid1[0:3, 0:3].position)

    def test12(self):
        grid1 = datamodel.new_regular_grid((4, 4), [1.0, 1.0] | units.m)
        grid2 = datamodel.new_regular_grid((4, 4), [1.0, 1.0] | units.m, offset=[-0.5, -0.5] | units.m)
        self.assertTrue(grid1.overlaps(grid2))
        overlap = grid1.get_overlap_with(grid2, eps=grid2.cellsize()[0]*1.e-6)
        self.assertEqual(overlap.position, grid1[0:2, 0:2].position)


class TestGridSampling(amusetest.TestCase):

    def test1(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertEqual(sample.index, [1, 1, 1])
        sample = grid.samplePoint([2.5, 2.5, 2.5] | units.m, method="interpolation")
        self.assertEqual(sample.index, [1, 1, 1])
        sample = grid.samplePoint([3.5, 3.5, 3.5] | units.m, method="interpolation")
        self.assertEqual(sample.index, [1, 1, 1])

        for x in range(0, 200):
            sample = grid.samplePoint([0.0 + (x/100.0), 4.0+(x/100.0), 6.0+(x/100.0)] | units.m, method="interpolation")
            self.assertEqual(sample.index, [0, 2, 3])

        for x in range(200, 400):
            sample = grid.samplePoint([0.0 + (x/100.0), 4.0+(x/100.0), 6.0+(x/100.0)] | units.m, method="interpolation")
            self.assertEqual(sample.index, [1, 3, 4])

    def test2(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertEqual(sample.index_for_000_cell, [1, 1, 1])
        sample = grid.samplePoint([2.5, 2.5, 2.5] | units.m, method="interpolation")
        self.assertEqual(sample.index_for_000_cell, [0, 0, 0])
        sample = grid.samplePoint([3.5, 3.5, 3.5] | units.m, method="interpolation")
        self.assertEqual(sample.index_for_000_cell, [1, 1, 1])
        sample = grid.samplePoint([4.5, 4.5, 4.5] | units.m, method="interpolation")
        self.assertEqual(sample.index_for_000_cell, [1, 1, 1])
        self.assertEqual(sample.index, [2, 2, 2])

        for x in range(0, 100):

            sample = grid.samplePoint([0.0 + (x/100.0), 4.0+(x/100.0), 6.0+(x/100.0)] | units.m, method="interpolation")
            self.assertEqual(sample.index_for_000_cell, [-1, 1, 2])
        for x in range(100, 300):

            sample = grid.samplePoint([0.0 + (x/100.0), 4.0+(x/100.0), 6.0+(x/100.0)] | units.m, method="interpolation")
            self.assertEqual(sample.index_for_000_cell, [0, 2, 3])

        for x in range(300, 400):
            sample = grid.samplePoint([0.0 + (x/100.0), 4.0+(x/100.0), 6.0+(x/100.0)] | units.m, method="interpolation")
            self.assertEqual(sample.index_for_000_cell, [1, 3, 4])

    def test3(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertEqual(sample.index_for_000_cell, [1, 1, 1])
        self.assertEqual(sample.surrounding_cell_indices, [
            [1, 1, 1],
            [2, 1, 1],
            [1, 2, 1],
            [1, 1, 2],
            [2, 1, 2],
            [1, 2, 2],
            [2, 2, 1],
            [2, 2, 2],
        ])

    def test4(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertEqual(sample.surrounding_cells[0].position, [3.0, 3.0, 3.0] | units.m)
        self.assertEqual(sample.surrounding_cells[1].position, [5.0, 3.0, 3.0] | units.m)
        self.assertEqual(sample.surrounding_cells[-1].position, [5.0, 5.0, 5.0] | units.m)

    def test5(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        masses = sample.get_values_of_attribute("mass")
        self.assertEqual(masses[0], 3.0 | units.kg)
        self.assertEqual(masses[1], 5.0 | units.kg)
        self.assertEqual(masses[2], 3.0 | units.kg)
        self.assertEqual(masses[3], 3.0 | units.kg)
        self.assertEqual(masses[4], 5.0 | units.kg)
        self.assertEqual(masses[5], 3.0 | units.kg)
        self.assertEqual(masses[6], 5.0 | units.kg)
        self.assertEqual(masses[7], 5.0 | units.kg)
        factors = sample.weighing_factors
        self.assertEqual(factors[0], 1.0 | units.none)
        self.assertEqual(factors[1], 0.0 | units.none)
        self.assertEqual(factors[2], 0.0 | units.none)
        self.assertEqual(factors[3], 0.0 | units.none)
        self.assertEqual(factors[4], 0.0 | units.none)
        self.assertEqual(factors[5], 0.0 | units.none)
        self.assertEqual(factors[6], 0.0 | units.none)
        self.assertEqual(factors[7], 0.0 | units.none)

        self.assertAlmostRelativeEquals(sample.mass, 3.0 | units.kg)

    def test6(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        for xpos in numpy.arange(3.0, 5.0, 0.1):
            sample = grid.samplePoint([xpos, 3.0, 3.0] | units.m, method="interpolation")
            self.assertAlmostRelativeEquals(sample.mass, (3.0 | units.kg) + ((2.0 * (xpos - 3.0) / 2.0) | units.kg))

            sample = grid.samplePoint([xpos, 3.0, 3.0] | units.m, method="interpolation")
            self.assertAlmostRelativeEquals(sample.mass, (3.0 | units.kg) + ((2.0 * (xpos - 3.0) / 2.0) | units.kg))

            sample = grid.samplePoint([xpos, 5.0, 3.0] | units.m, method="interpolation")
            self.assertAlmostRelativeEquals(sample.mass, (3.0 | units.kg) + ((2.0 * (xpos - 3.0) / 2.0) | units.kg))

            sample = grid.samplePoint([xpos, 3.0, 5.0] | units.m, method="interpolation")
            self.assertAlmostRelativeEquals(sample.mass, (3.0 | units.kg) + ((2.0 * (xpos - 3.0) / 2.0) | units.kg))

            sample = grid.samplePoint([4.0, 4.0, 4.0] | units.m, method="interpolation")
            self.assertAlmostRelativeEquals(sample.mass, (4.0 | units.kg))

    def test7(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertTrue(sample.isvalid)
        sample = grid.samplePoint([11.0, 3.0, 3.0] | units.m, method="interpolation")
        self.assertFalse(sample.isvalid)
        sample = grid.samplePoint([3.0, -1.0, 3.0] | units.m, method="interpolation")
        self.assertFalse(sample.isvalid)

    def test8(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="nearest")
        self.assertEqual(sample.position, [3.0, 3.0, 3.0] | units.m)
        self.assertEqual(sample.mass, 3.0 | units.kg)
        sample = grid.samplePoint([3.5, 3.0, 3.0] | units.m, method="nearest")
        self.assertEqual(sample.position, [3.0, 3.0, 3.0] | units.m)
        self.assertEqual(sample.mass, 3.0 | units.kg)

    def test9(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        sample = grid.samplePoint([3.0, 3.0, 3.0] | units.m, method="linear")
        self.assertEqual(sample.position, [3.0, 3.0, 3.0] | units.m)
        self.assertEqual(sample.mass, 3.0 | units.kg)
        sample = grid.samplePoint([3.5, 3.0, 3.0] | units.m, method="linear")
        self.assertEqual(sample.position, [3.5, 3.0, 3.0] | units.m)
        self.assertEqual(sample.mass, 3.5 | units.kg)


class TestGridSamplingMultiplePoints(amusetest.TestCase):

    def test1(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        samples = grid.samplePoints([[3.0, 3.0, 3.0], [4.0, 3.0, 3.0]] | units.m, method="linear")
        self.assertEqual(len(samples), 2)
        self.assertEqual(samples.position[0], [3.0, 3.0, 3.0] | units.m)
        self.assertEqual(samples.position[0], samples[0].position)
        self.assertEqual(samples.position[1], samples[1].position)
        self.assertEqual(samples.mass, [3.0, 4.0] | units.kg)

    def test2(self):
        grid = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid.mass = grid.x.value_in(units.m) | units.kg
        samples = grid.samplePoints([[3.5, 3.0, 3.0], [4.5, 3.0, 3.0]] | units.m, method="linear")
        self.assertEqual(len(samples), 2)
        self.assertEqual(samples.mass, [3.5, 4.5] | units.kg)

    def test3(self):
        grid1 = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid1.mass = grid1.x.value_in(units.m) | units.kg
        grid2 = datamodel.new_regular_grid((5, 5, 5), [10.0, 10.0, 10.0] | units.m)
        grid2.position += (10.0, 0, 0) | units.m
        grid2.mass = grid2.x.value_in(units.m) | units.kg
        samples = SamplePointsOnMultipleGrids((grid1, grid2), [[3.0, 3.0, 3.0], [4.0, 3.0, 3.0], [13, 3, 3]] | units.m)
        self.assertEqual(len(samples), 3)
        self.assertEqual(samples.mass, [3.0, 4.0, 13.0] | units.kg)
